# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file has functions to create NURBS geometries using Geomdl."""

import numpy as _np
from geomdl import NURBS as _NURBS
from geomdl import compatibility as _compat
from geomdl import operations as _operations


def create_nurbs_hollow_cylinder_segment_2d(
    radius_in, radius_out, angle, *, n_ele_u=1, n_ele_v=1
):
    """Creates a patch of a 2 dimensional segment of a hollow cylinder.

    Args
    ----
    radius_in: double
        inner cylinder radius
    radius_out: double
        outer cylinder radius
    angle: double
        angle of the hollow cylinder section (radians)
    n_ele_u: int
        number of elements in the parametric u-direction
    n_ele_v: int
        number of elements in the parametric v-direction


    Return
    ----
    surf: geomdl object
        geomdl object that contains the surface information
    """

    # Check the validity of the input values:
    if radius_in >= radius_out:
        raise ValueError(
            "The external radius should be larger than the internal radius of a hollow cylinder."
        )
    if (angle > _np.pi) or (angle < 0):
        raise ValueError(
            "The following algorithm for creating a hollow cylinder section is only valid for 0 < angle <= pi."
        )

    # Create a NURBS surface instance
    surf = _NURBS.Surface()

    # Set degrees
    surf.degree_u = 2
    surf.degree_v = 2

    # Control points and set them to the surface
    p_size_u = 3
    p_size_v = 3

    # To calculate the internal control point cp_2 we have to calculate
    # the intersection of the tangents of the points cp_1 and cp_3.
    # As point cp_1 is always defined over the x axis, its tangent is simply x = radius
    # The equation of the tangent to the circle at the cp_3 point is:
    # y - cp_3y = -(cp_3x/cp_3y)*(x - cp_3x).

    # Obtaining the control points that define the external arc of the hollow cylinder
    cp_ext1 = [radius_out, 0.0, 0.0]
    cp_ext3 = [radius_out * _np.cos(angle), radius_out * _np.sin(angle), 0.0]
    cp_ext2 = [
        radius_out,
        -(cp_ext3[0] / cp_ext3[1]) * (radius_out - cp_ext3[0]) + cp_ext3[1],
        0.0,
    ]

    # Obtaining the control points that define the internal arc of the hollow cylinder
    cp_int1 = [radius_in, 0.0, 0.0]
    cp_int3 = [radius_in * _np.cos(angle), radius_in * _np.sin(angle), 0.0]
    cp_int2 = [
        radius_in,
        -(cp_int3[0] / cp_int3[1]) * (radius_in - cp_int3[0]) + cp_int3[1],
        0.0,
    ]

    # Obtaining the control points positioned in the middle of the internal and external arches
    cp_middle1 = [(cp_ext1[0] + cp_int1[0]) / 2, (cp_ext1[1] + cp_int1[1]) / 2, 0.0]
    cp_middle2 = [(cp_ext2[0] + cp_int2[0]) / 2, (cp_ext2[1] + cp_int2[1]) / 2, 0.0]
    cp_middle3 = [(cp_ext3[0] + cp_int3[0]) / 2, (cp_ext3[1] + cp_int3[1]) / 2, 0.0]

    ctrlpts = [
        cp_ext1,
        cp_ext2,
        cp_ext3,
        cp_middle1,
        cp_middle2,
        cp_middle3,
        cp_int1,
        cp_int2,
        cp_int3,
    ]

    weights = [
        1.0,
        _np.cos(angle / 2),
        1.0,
        1.0,
        _np.cos(angle / 2),
        1.0,
        1.0,
        _np.cos(angle / 2),
        1.0,
    ]

    t_ctrlptsw = _compat.combine_ctrlpts_weights(ctrlpts, weights)
    n_ctrlptsw = _compat.flip_ctrlpts_u(t_ctrlptsw, p_size_u, p_size_v)

    surf.ctrlpts_size_u = p_size_u
    surf.ctrlpts_size_v = p_size_v
    surf.ctrlptsw = n_ctrlptsw

    surf.knotvector_u = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]
    surf.knotvector_v = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]

    do_uniform_knot_refinement_surface(surf, n_ele_u, n_ele_v)

    return surf


def create_nurbs_cylindrical_shell_sector(
    radius: float, angle: float, length: float, *, n_ele_u: int = 1, n_ele_v: int = 1
) -> _NURBS.Surface:
    """Creates a NURBS surface representing a 3D sector of a cylindrical shell.
    The center of the cylindrical shell sector is located at [0, 0, 0].

    Args:
        radius: Radius of the cylindrical shell.
        angle: Angle of the cylindrical shell sector in radians.
                       The angle is only valid for 0 < angle <= pi according
                       to "Isogeometric Analysis: Toward Integration of CAD and FEA" by
                       J.Austin Cottrell, 2009
        length: Length of the cylindrical shell.
        n_ele_u: Number of elements in the parametric u-direction. Defaults to 1.
        n_ele_v: Number of elements in the parametric v-direction. Defaults to 1.

    Returns:
        geomdl.NURBS.Surface: A geomdl object that contains the surface information.
    """

    # Check the validity of the input values:
    if (angle >= _np.pi) or (angle < 0):
        raise ValueError(
            "The following algorithm for creating a cylindrical shell sector is only valid for 0 < angle <= pi."
        )

    # Create a NURBS surface instance
    surf = _NURBS.Surface()

    # Set degrees
    surf.degree_u = 2
    surf.degree_v = 2

    # Control points and set them to the surface
    p_size_u = 3
    p_size_v = 3

    # Obtaining the control points
    cp_1 = [-radius * _np.sin(angle / 2), -length / 2, -radius * _np.cos(angle / 2)]
    cp_3 = [radius * _np.sin(angle / 2), -length / 2, -radius * _np.cos(angle / 2)]

    # Calculating position of middle points. This is done by
    # obtaining the tangents on the points cp_1 and cp_3 and
    # calculating the intersection point between the tangents.
    m_1 = -_np.tan(-angle / 2)
    m_3 = -_np.tan(angle / 2)
    b_1 = cp_1[2] + m_1 * cp_1[0]
    b_3 = cp_3[2] + m_3 * cp_3[0]

    inter_point_x = (b_3 - b_1) / (m_3 - m_1)
    inter_point_z = m_1 * inter_point_x + b_1

    # The intersection point is assigned to the middle points cp_2, cp_5 and cp_8
    cp_2 = [inter_point_x, -length / 2, inter_point_z]

    cp_4 = [-radius * _np.sin(angle / 2), 0.0, -radius * _np.cos(angle / 2)]
    cp_5 = [inter_point_x, 0.0, inter_point_z]
    cp_6 = [radius * _np.sin(angle / 2), 0.0, -radius * _np.cos(angle / 2)]

    cp_7 = [
        -radius * _np.sin(angle / 2),
        length / 2,
        -radius * _np.cos(angle / 2),
    ]
    cp_8 = [inter_point_x, length / 2, inter_point_z]
    cp_9 = [radius * _np.sin(angle / 2), length / 2, -radius * _np.cos(angle / 2)]

    ctrlpts = [cp_1, cp_4, cp_7, cp_2, cp_5, cp_8, cp_3, cp_6, cp_9]

    weights = [
        1.0,
        1.0,
        1.0,
        _np.cos(angle / 2),
        _np.cos(angle / 2),
        _np.cos(angle / 2),
        1.0,
        1.0,
        1.0,
    ]

    t_ctrlptsw = _compat.combine_ctrlpts_weights(ctrlpts, weights)

    surf.ctrlpts_size_u = p_size_u
    surf.ctrlpts_size_v = p_size_v
    surf.ctrlptsw = t_ctrlptsw

    surf.knotvector_u = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]
    surf.knotvector_v = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]

    do_uniform_knot_refinement_surface(surf, n_ele_u, n_ele_v)

    return surf


def create_nurbs_flat_plate_2d(width, length, *, n_ele_u=1, n_ele_v=1):
    """Creates a patch of a 2 dimensional flat plate.

    Args
    ----
    width: double
        dimension of the plate in the x-direction
    length: double
        dimension of the plate in the y-direction
    n_ele_u: int
        number of elements in the parametric u-direction
    n_ele_v: int
        number of elements in the parametric v-direction


    Return
    ----
    surf: geomdl object
        geomdl object that contains the surface information
    """

    # Create a NURBS surface instance
    surf = _NURBS.Surface()

    # Set degrees
    surf.degree_u = 2
    surf.degree_v = 2

    # Control points and set them to the surface
    p_size_u = 3
    p_size_v = 3

    ctrlpts = [
        [-width / 2, -length / 2, 0.0],
        [0.0, -length / 2, 0.0],
        [width / 2, -length / 2, 0.0],
        [-width / 2, 0.0, 0.0],
        [0.0, 0.0, 0.0],
        [width / 2, 0.0, 0.0],
        [-width / 2, length / 2, 0.0],
        [0.0, length / 2, 0.0],
        [width / 2, length / 2, 0.0],
    ]

    weights = [1.0] * 9

    t_ctrlptsw = _compat.combine_ctrlpts_weights(ctrlpts, weights)
    n_ctrlptsw = _compat.flip_ctrlpts_u(t_ctrlptsw, p_size_u, p_size_v)

    surf.ctrlpts_size_u = p_size_u
    surf.ctrlpts_size_v = p_size_v
    surf.ctrlptsw = n_ctrlptsw

    surf.knotvector_u = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]
    surf.knotvector_v = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]

    do_uniform_knot_refinement_surface(surf, n_ele_u, n_ele_v)

    return surf


def create_nurbs_sphere_surface(radius, n_ele_u=1, n_ele_v=1):
    """Generates a patch of a sphere as a NURBS surface. This function
    constructs a segment of a spherical surface using Non-Uniform Rational
    B-Splines (NURBS) based on the specified radius and the number of elements
    in the parametric u and v directions.

    Args
    ---
    radius: double
        radius of the sphere
    n_ele_u: int
        number of elements in the parametric u-direction
    n_ele_v: int
        number of elements in the parametric v-direction

    Return
    ----
    surf: geomdl object
        geomdl object that contains the surface information
    """

    # Create a NURBS surface instance
    surf = _NURBS.Surface()

    # Set degrees
    surf.degree_u = 2
    surf.degree_v = 2

    # Control points and set them to the surface
    p_size_u = 3
    p_size_v = 3

    dummy = 6.0 * (
        5.0 / 12.0
        + 0.5 * _np.sqrt(2.0 / 3.0)
        - 0.25 / _np.sqrt(3.0)
        - 0.5 * _np.sqrt(2.0 / 3.0) * _np.sqrt(3.0) / 2.0
    )

    ctrlpts = [
        [
            2.0 * radius / _np.sqrt(6.0) * -_np.sin(1.0 / 4.0 * _np.pi),
            -radius / _np.sqrt(3.0),
            2.0 * radius / _np.sqrt(6.0) * _np.cos(1.0 / 4.0 * _np.pi),
        ],
        [
            radius * _np.sqrt(3.0) / (_np.sqrt(2.0) * 2.0) * _np.cos(1.0 / 4.0 * _np.pi)
            + radius
            * _np.sqrt(3.0)
            / (_np.sqrt(2.0) * 2.0)
            * -_np.sin(1.0 / 4.0 * _np.pi),
            -_np.sqrt(3.0) / 2 * radius,
            radius * _np.sqrt(3.0) / (_np.sqrt(2.0) * 2.0) * _np.cos(1.0 / 4.0 * _np.pi)
            + radius
            * _np.sqrt(3.0)
            / (_np.sqrt(2.0) * 2.0)
            * _np.sin(1.0 / 4.0 * _np.pi),
        ],
        [
            2.0 * radius / _np.sqrt(6.0) * _np.cos(1.0 / 4.0 * _np.pi),
            -radius / _np.sqrt(3.0),
            2.0 * radius / _np.sqrt(6.0) * _np.sin(1.0 / 4.0 * _np.pi),
        ],
        [
            radius * _np.sqrt(6.0) / 2.0 * -_np.sin(1.0 / 4.0 * _np.pi),
            0.0,
            radius * _np.sqrt(6.0) / 2.0 * _np.cos(1.0 / 4.0 * _np.pi),
        ],
        [
            radius * dummy * _np.sqrt(2.0) / 2.0 * _np.cos(1.0 / 4.0 * _np.pi)
            + radius * dummy * _np.sqrt(2.0) / 2.0 * -_np.sin(1.0 / 4.0 * _np.pi),
            0.0,
            radius * dummy * _np.sqrt(2.0) / 2.0 * _np.cos(1.0 / 4.0 * _np.pi)
            + radius * dummy * _np.sqrt(2.0) / 2.0 * _np.sin(1.0 / 4.0 * _np.pi),
        ],
        [
            radius * _np.sqrt(6.0) / 2.0 * _np.cos(1.0 / 4.0 * _np.pi),
            0.0,
            radius * _np.sqrt(6.0) / 2.0 * _np.sin(1.0 / 4.0 * _np.pi),
        ],
        [
            2.0 * radius / _np.sqrt(6.0) * -_np.sin(1.0 / 4.0 * _np.pi),
            2.0 * radius / _np.sqrt(6.0) * _np.cos(1.0 / 4.0 * _np.pi),
            radius / _np.sqrt(3.0),
        ],
        [
            radius * _np.sqrt(3.0) / (_np.sqrt(2.0) * 2.0) * _np.cos(1.0 / 4.0 * _np.pi)
            + radius
            * _np.sqrt(3.0)
            / (_np.sqrt(2.0) * 2.0)
            * -_np.sin(1.0 / 4.0 * _np.pi),
            _np.sqrt(3.0) / 2 * radius,
            radius * _np.sqrt(3.0) / (_np.sqrt(2.0) * 2.0) * _np.cos(1.0 / 4.0 * _np.pi)
            + radius
            * _np.sqrt(3.0)
            / (_np.sqrt(2.0) * 2.0)
            * _np.sin(1.0 / 4.0 * _np.pi),
        ],
        [
            2.0 * radius / _np.sqrt(6.0) * _np.cos(1.0 / 4.0 * _np.pi),
            radius / _np.sqrt(3.0),
            2.0 * radius / _np.sqrt(6.0) * _np.sin(1.0 / 4.0 * _np.pi),
        ],
    ]

    weights = [
        1.0,
        2.0 / _np.sqrt(6.0),
        1.0,
        2.0 / _np.sqrt(6.0),
        2.0 / 3.0,
        2.0 / _np.sqrt(6.0),
        1.0,
        2.0 / _np.sqrt(6.0),
        1.0,
    ]

    t_ctrlptsw = _compat.combine_ctrlpts_weights(ctrlpts, weights)
    n_ctrlptsw = _compat.flip_ctrlpts_u(t_ctrlptsw, p_size_u, p_size_v)

    surf.ctrlpts_size_u = p_size_u
    surf.ctrlpts_size_v = p_size_v
    surf.ctrlptsw = n_ctrlptsw

    surf.knotvector_u = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]
    surf.knotvector_v = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]

    do_uniform_knot_refinement_surface(surf, n_ele_u, n_ele_v)

    return surf


def create_nurbs_hemisphere_surface(radius, n_ele_uv=1):
    """Generates a hemisphere as a NURBS surface. This function constructs five
    segments that represent the surface of a hemisphere using Non-Uniform
    Rational B-Splines (NURBS) based on the specified radius and the number of
    elements in the parametric u and v directions. To secure the connectivity
    between surfaces, all surfaces must have the same parametric representation
    in any parametric direction. Therefore, the number of elements in u- and v-
    directions must be the same.

    This function generates a list of five NURBS geomdl objects.

    Args
    ---
    radius: double
        Radius of the hemisphere
    n_ele_uv: int
        Number of elements in the parametric u- and v- directions

    Return
    ----
    list: list(geomdl object)
        A list of geomdl objects that contains the surface information
    """

    # Create the first section of the hemisphere
    hemisphere_1 = create_nurbs_sphere_surface(radius, n_ele_u=1, n_ele_v=1)

    # Create a temporary section by rotating and translating the
    # first section of the hemisphere
    temp_hemisphere = _operations.rotate(hemisphere_1, 90, axis=1)
    temp_hemisphere = _operations.translate(
        temp_hemisphere,
        (0, 0, -2.0 * radius / _np.sqrt(6.0) * _np.sin(1.0 / 4.0 * _np.pi) * 2),
    )

    # To create the hemisphere it is necessary to split the temporary
    # sphere section in two pieces. This split is done in u = 0.5. After
    # the split, the second surface is taken as it will be
    # adjacent to the first section of the sphere
    cut_section_sphere = _operations.split_surface_u(temp_hemisphere, param=0.5)
    hemisphere_2 = cut_section_sphere[1]

    translation_component = radius * _np.sin(1.0 / 4.0 * _np.pi) * 2
    # Create the third section. Rotate and translate it accordingly
    hemisphere_3 = _operations.rotate(hemisphere_2, 90, axis=2)
    hemisphere_3 = _operations.translate(hemisphere_3, (translation_component, 0, 0))

    # Create the forth section. Rotate and translate it accordingly
    hemisphere_4 = _operations.rotate(hemisphere_3, 90, axis=2)
    hemisphere_4 = _operations.translate(hemisphere_4, (0, translation_component, 0))

    # Create the fifth section. Rotate and translate it accordingly
    hemisphere_5 = _operations.rotate(hemisphere_4, 90, axis=2)
    hemisphere_5 = _operations.translate(hemisphere_5, (-translation_component, 0, 0))

    patches = [hemisphere_1, hemisphere_2, hemisphere_3, hemisphere_4, hemisphere_5]
    for patch in patches:
        do_uniform_knot_refinement_surface(patch, n_ele_uv, n_ele_uv)
    return patches


def create_nurbs_torus_surface(radius_torus, radius_circle, *, n_ele_u=1, n_ele_v=1):
    """Creates a NURBS patch for the construction of a ring torus (outer
    surface). A ring torus is a surface of revolution generated by revolving a
    circle in a 3-dimensional space around an axis of revolution (axis z). The
    center of the torus is located at the coordinates [0, 0, 0].

    This function constructs the surface of a ring torus in the following manner:
     - The quarter of the torus is made by 4 patches
     - These patches are rotated three times around the z-axis by 90°, e.g. theta = 90°, 180°, 270°,
       leading to a total of 16 NURBS patches.

    The number of elements in this function are given for a single patch, where:
     - n_ele_u is the number of elements in the "radius_torus" direction
     - n_ele_v is the number of elements in the "radius_circle" direction

    This function generates a list of 16 NURBS geomdl objects.

    Args
    ----
    radius_torus: double
        distance from the axis of revolution (axis z) to the center of the revolving circle
    radius_circle: double
        radius of the circle that revolves around the axis of revolution
    n_ele_u: int
        number of elements in the parametric u-direction
    n_ele_v: int
        number of elements in the parametric v-direction
    """

    # Create four NURBS surface instances. These are the base patches of the torus.
    surf_1 = _NURBS.Surface()
    surf_2 = _NURBS.Surface()
    surf_3 = _NURBS.Surface()
    surf_4 = _NURBS.Surface()
    base_surfs = [surf_1, surf_2, surf_3, surf_4]

    # Define control points and set them to the surfaces
    p_size_u = 3
    p_size_v = 3

    dummy_surf1 = radius_torus + radius_circle
    dummy_surf2 = radius_torus - radius_circle

    ctrlpts_surf1 = [
        [dummy_surf1, 0.0, 0.0],
        [dummy_surf1, 0.0, radius_circle],
        [radius_torus, 0.0, radius_circle],
        [dummy_surf1, dummy_surf1, 0.0],
        [dummy_surf1, dummy_surf1, radius_circle],
        [radius_torus, radius_torus, radius_circle],
        [0.0, dummy_surf1, 0.0],
        [0.0, dummy_surf1, radius_circle],
        [0.0, radius_torus, radius_circle],
    ]

    ctrlpts_surf2 = [
        [dummy_surf1, 0.0, 0.0],
        [dummy_surf1, 0.0, -radius_circle],
        [radius_torus, 0.0, -radius_circle],
        [dummy_surf1, dummy_surf1, 0.0],
        [dummy_surf1, dummy_surf1, -radius_circle],
        [radius_torus, radius_torus, -radius_circle],
        [0.0, dummy_surf1, 0.0],
        [0.0, dummy_surf1, -radius_circle],
        [0.0, radius_torus, -radius_circle],
    ]

    ctrlpts_surf3 = [
        [radius_torus, 0.0, -radius_circle],
        [dummy_surf2, 0.0, -radius_circle],
        [dummy_surf2, 0.0, 0.0],
        [radius_torus, radius_torus, -radius_circle],
        [dummy_surf2, dummy_surf2, -radius_circle],
        [dummy_surf2, dummy_surf2, 0.0],
        [0.0, radius_torus, -radius_circle],
        [0.0, dummy_surf2, -radius_circle],
        [0.0, dummy_surf2, 0.0],
    ]

    ctrlpts_surf4 = [
        [0.0, dummy_surf2, 0.0],
        [0.0, dummy_surf2, radius_circle],
        [0.0, radius_torus, radius_circle],
        [dummy_surf2, dummy_surf2, 0.0],
        [dummy_surf2, dummy_surf2, radius_circle],
        [radius_torus, radius_torus, radius_circle],
        [dummy_surf2, 0.0, 0.0],
        [dummy_surf2, 0.0, radius_circle],
        [radius_torus, 0.0, radius_circle],
    ]

    weights = [
        1.0,
        _np.sqrt(2) / 2,
        1.0,
        _np.sqrt(2) / 2,
        0.5,
        _np.sqrt(2) / 2,
        1.0,
        _np.sqrt(2) / 2,
        1.0,
    ]

    t_ctrlptsw1 = _compat.combine_ctrlpts_weights(ctrlpts_surf1, weights)
    t_ctrlptsw2 = _compat.combine_ctrlpts_weights(ctrlpts_surf2, weights)
    t_ctrlptsw3 = _compat.combine_ctrlpts_weights(ctrlpts_surf3, weights)
    t_ctrlptsw4 = _compat.combine_ctrlpts_weights(ctrlpts_surf4, weights)

    t_ctrlpts_surfs = [t_ctrlptsw1, t_ctrlptsw2, t_ctrlptsw3, t_ctrlptsw4]

    for surf, t_ctrlpts in zip(base_surfs, t_ctrlpts_surfs):
        # Set degrees
        surf.degree_u = 2
        surf.degree_v = 2

        surf.ctrlpts_size_u = p_size_u
        surf.ctrlpts_size_v = p_size_v

        # Set control points
        surf.ctrlptsw = t_ctrlpts

        # Set knot vectors
        surf.knotvector_u = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]
        surf.knotvector_v = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]

        do_uniform_knot_refinement_surface(surf, n_ele_u, n_ele_v)

    # Define the rotations and translations to rotate the base patches and form a complete torus
    tmp_trans = [
        radius_torus + radius_circle,
        radius_torus + radius_circle,
        radius_torus,
        radius_torus - radius_circle,
    ]

    transform_surf1 = [
        [(-tmp_trans[0], tmp_trans[0], 0), 90, 2],
        [(-2 * tmp_trans[0], 0, 0), 180, 2],
        [(-tmp_trans[0], -tmp_trans[0], 0), 270, 2],
    ]

    transform_surf2 = [
        [(-tmp_trans[1], tmp_trans[1], 0), 90, 2],
        [(-2 * tmp_trans[1], 0, 0), 180, 2],
        [(-tmp_trans[1], -tmp_trans[1], 0), 270, 2],
    ]

    transform_surf3 = [
        [(-tmp_trans[2], tmp_trans[2], 0), 90, 2],
        [(-2 * tmp_trans[2], 0, 0), 180, 2],
        [(-tmp_trans[2], -tmp_trans[2], 0), 270, 2],
    ]

    transform_surf4 = [
        [(-tmp_trans[3], -tmp_trans[3], 0), 90, 2],
        [(0, -2 * tmp_trans[3], 0), 180, 2],
        [(tmp_trans[3], -tmp_trans[3], 0), 270, 2],
    ]

    # Rotate base patches and store them
    surfaces_torus = [surf_1, surf_2, surf_3, surf_4]
    for transform1, transform2, transform3, transform4 in zip(
        transform_surf1, transform_surf2, transform_surf3, transform_surf4
    ):
        new_surf1 = _operations.translate(surf_1, transform1[0])
        new_surf1 = _operations.rotate(new_surf1, transform1[1], axis=transform1[2])
        surfaces_torus.append(new_surf1)

        new_surf2 = _operations.translate(surf_2, transform2[0])
        new_surf2 = _operations.rotate(new_surf2, transform2[1], axis=transform2[2])
        surfaces_torus.append(new_surf2)

        new_surf3 = _operations.translate(surf_3, transform3[0])
        new_surf3 = _operations.rotate(new_surf3, transform3[1], axis=transform3[2])
        surfaces_torus.append(new_surf3)

        new_surf4 = _operations.translate(surf_4, transform4[0])
        new_surf4 = _operations.rotate(new_surf4, transform4[1], axis=transform4[2])
        surfaces_torus.append(new_surf4)

    return surfaces_torus


def create_nurbs_brick(width, length, height, *, n_ele_u=1, n_ele_v=1, n_ele_w=1):
    """Creates a patch of a 3 dimensional brick.

    Args
    ----
    width: double
        dimension of the plate in the x-direction
    length: double
        dimension of the plate in the y-direction
    height: double
        dimension of the plate in the z-direction
    n_ele_u: int
        number of elements in the parametric u-direction
    n_ele_v: int
        number of elements in the parametric v-direction
    n_ele_w: int
        number of elements in the parametric w-direction


    Return
    ----
    vol: geomdl object
        geomdl object that contains the volume information
    """

    # Create a NURBS volume instance
    vol = _NURBS.Volume()

    # Set degrees
    vol.degree_u = 2
    vol.degree_v = 2
    vol.degree_w = 2

    # Create control points and set them to the volume
    cp_size_u = 3
    cp_size_v = 3
    cp_size_w = 3

    ctrlpts = [
        [-width / 2, -length / 2, -height / 2],
        [-width / 2, 0.0, -height / 2],
        [-width / 2, length / 2, -height / 2],
        [0.0, -length / 2, -height / 2],
        [0.0, 0.0, -height / 2],
        [0.0, length / 2, -height / 2],
        [width / 2, -length / 2, -height / 2],
        [width / 2, 0.0, -height / 2],
        [width / 2, length / 2, -height / 2],
        [-width / 2, -length / 2, 0.0],
        [-width / 2, 0.0, 0.0],
        [-width / 2, length / 2, 0.0],
        [0.0, -length / 2, 0.0],
        [0.0, 0.0, 0.0],
        [0.0, length / 2, 0.0],
        [width / 2, -length / 2, 0.0],
        [width / 2, 0.0, 0.0],
        [width / 2, length / 2, 0.0],
        [-width / 2, -length / 2, height / 2],
        [-width / 2, 0.0, height / 2],
        [-width / 2, length / 2, height / 2],
        [0.0, -length / 2, height / 2],
        [0.0, 0.0, height / 2],
        [0.0, length / 2, height / 2],
        [width / 2, -length / 2, height / 2],
        [width / 2, 0.0, height / 2],
        [width / 2, length / 2, height / 2],
    ]

    weights = [1.0] * 27

    vol.ctrlpts_size_u = cp_size_u
    vol.ctrlpts_size_v = cp_size_v
    vol.ctrlpts_size_w = cp_size_w

    t_ctrlptsw = _compat.combine_ctrlpts_weights(ctrlpts, weights)

    vol.set_ctrlpts(t_ctrlptsw, cp_size_u, cp_size_v, cp_size_w)

    vol.knotvector_u = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]
    vol.knotvector_v = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]
    vol.knotvector_w = [0.0, 0.0, 0.0, 1.0, 1.0, 1.0]

    do_uniform_knot_refinement_volume(vol, n_ele_u, n_ele_v, n_ele_w)

    return vol


def do_uniform_knot_refinement_surface(surf, n_ele_u, n_ele_v):
    """This function does an uniform knot refinement in the u- and v-
    direction.

    Args
    ----
    surf: geomdl object
        geomdl object that contains the surface information
    n_ele_u: int
        number of elements in the parametric u-direction
    n_ele_v: int
        number of elements in the parametric v-direction

    Return
    ----
    surf: geomdl object
    """

    size_of_knotvector_u = 1 / n_ele_u
    size_of_knotvector_v = 1 / n_ele_v

    for i in range(1, n_ele_u):
        _operations.insert_knot(surf, [size_of_knotvector_u * i, None], [1, 0])
    for j in range(1, n_ele_v):
        _operations.insert_knot(surf, [None, size_of_knotvector_v * j], [0, 1])


def do_uniform_knot_refinement_volume(vol, n_ele_u, n_ele_v, n_ele_w):
    """This function does an uniform knot refinement in the u-, v- and w-
    direction.

    Args
    ----
    vol: geomdl object
        geomdl object that contains the volume information
    n_ele_u: int
        number of elements in the parametric u-direction
    n_ele_v: int
        number of elements in the parametric v-direction
    n_ele_w: int
        number of elements in the parametric w-direction

    Return
    ----
    vol: geomdl object
    """

    size_of_knotvector_u = 1 / n_ele_u
    size_of_knotvector_v = 1 / n_ele_v
    size_of_knotvector_w = 1 / n_ele_w

    for i in range(1, n_ele_u):
        _operations.insert_knot(vol, [size_of_knotvector_u * i, None, None], [1, 0, 0])
    for j in range(1, n_ele_v):
        _operations.insert_knot(vol, [None, size_of_knotvector_v * j, None], [0, 1, 0])
    for k in range(1, n_ele_w):
        _operations.insert_knot(vol, [None, None, size_of_knotvector_w * k], [0, 0, 1])



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module provides a class that is used to write VTK files."""

import numbers as _numbers
import os as _os
import warnings as _warnings
from enum import Enum as _Enum
from enum import auto as _auto

import numpy as _np
import vtk as _vtk

from beamme.core.conf import bme as _bme

# Number of digits for node set output (this will be set in the
# Mesh.get_unique_geometry_sets() method.
VTK_NODE_SET_FORMAT = "{:05}"

# Nan values for vtk data, since we currently can't set nan explicitly.
VTK_NAN_INT = -1
VTK_NAN_FLOAT = 0.0


class VTKGeometry(_Enum):
    """Enum for VTK geometry types (for now cells and points)."""

    point = _auto()
    cell = _auto()


class VTKType(_Enum):
    """Enum for VTK value types."""

    int = _auto()
    float = _auto()


class VTKTensor(_Enum):
    """Enum for VTK tensor types."""

    scalar = _auto()
    vector = _auto()


def add_point_data_node_sets(point_data, nodes, *, extra_points=0):
    """Add the information if a node is part of a set to the point_data vector
    for all nodes in the list 'nodes'.

    The extra_points argument specifies how many additional
    visualization points there are, i.e., points that are not based on
    nodes, but are only used for visualization purposes.
    """

    # Get list with node set indices of the given nodes
    geometry_set_list = []
    for node in nodes:
        geometry_set_list.extend(node.node_sets_link)

    # Remove double entries of list.
    geometry_set_list = list(set(geometry_set_list))

    # Loop over the geometry sets.
    n_nodes = len(nodes)
    for geometry_set in geometry_set_list:
        # Check which nodes are connected to a geometry set.
        data_vector = _np.zeros(n_nodes + extra_points)
        for i, node in enumerate(nodes):
            if geometry_set in node.node_sets_link:
                data_vector[i] = 1
            else:
                data_vector[i] = VTK_NAN_INT
        for i in range(extra_points):
            data_vector[n_nodes + i] = (
                1 if geometry_set.geometry_type is _bme.geo.line else VTK_NAN_INT
            )

        # Get the name of the geometry type.
        if geometry_set.geometry_type is _bme.geo.point:
            geometry_name = "geometry_point"
        elif geometry_set.geometry_type is _bme.geo.line:
            geometry_name = "geometry_line"
        elif geometry_set.geometry_type is _bme.geo.surface:
            geometry_name = "geometry_surface"
        elif geometry_set.geometry_type is _bme.geo.volume:
            geometry_name = "geometry_volume"
        else:
            raise TypeError("The geometry type is wrong!")

        # Add the data vector.
        set_name = f"{geometry_name}_set_{_bme.vtk_node_set_format.format(geometry_set.i_global + 1)}"
        point_data[set_name] = (data_vector, VTKType.int)


def _get_data_value_and_type(data):
    """Return the data and its type if one was given.

    The default type, if none was given is float.
    """
    if isinstance(data, tuple):
        return data[0], data[1]
    else:
        return data, VTKType.float


def _get_vtk_array_type(data):
    """Return the corresponding beamme type."""
    data_type = data.GetDataTypeAsString()
    if data_type == "int":
        return VTKType.int
    elif data_type == "double":
        return VTKType.float
    raise ValueError(f'Got unexpected type "{data_type}"!')


class VTKWriter:
    """A class that manages VTK cells and data and can also create them."""

    def __init__(self):
        # Initialize VTK objects.
        self.points = _vtk.vtkPoints()
        self.points.SetDataTypeToDouble()
        self.grid = _vtk.vtkUnstructuredGrid()

        # Link points to grid.
        self.grid.SetPoints(self.points)

        # Container for output data.
        self.data = {}
        for key1 in VTKGeometry:
            for key2 in VTKTensor:
                self.data[key1, key2] = {}

    def add_points(self, points, *, point_data=None):
        """Add points to the data stored in this object.

        Args
        ----
        points: [3d vector]
            Coordinates of points for this cell.
        point_data: dic
            A dictionary containing data that will be added for the newly added points.
            If a field exists in the global data but not in the one added here, that field
            will be set to bme.vtk_nan for the newly added points.

        Return:
        ----
        indices: [int]
            A list with the global indices of the added points.
        """

        n_points = len(points)

        # Check if point data containers are of the correct size
        if point_data is not None:
            for key, item_value in point_data.items():
                value, _data_type = _get_data_value_and_type(item_value)
                if not len(value) == n_points:
                    raise IndexError(
                        f"The length of coordinates is {n_points},"
                        f"the length of {key} is {len(value)}, does not match!"
                    )

        # Add point data
        self._add_data(point_data, VTKGeometry.point, n_new_items=n_points)

        # Add point coordinates
        n_grid_points = self.points.GetNumberOfPoints()
        for point in points:
            # Add the coordinate to the global list of coordinates.
            self.points.InsertNextPoint(*point)

        return _np.array(
            [n_grid_points + i_point for i_point in range(len(points))], dtype=int
        )

    def add_cell(self, cell_type, topology, *, cell_data=None):
        """Create a cell and add it to the global array.

        Args
        ----
        cell_type: VTK_type
            Type of cell that will be created.
        topology: [int]
            The connectivity between the cell and the global points.
        cell_data: dic
            A dictionary containing data that will be added for the newly added cell.
            If a field exists in the global data but not in the one added here, that field
            will be set to bme.vtk_nan for the newly added cell.
        """

        # Add the data entries.
        self._add_data(cell_data, VTKGeometry.cell)

        # Create the cell.
        geometry_item = cell_type()
        geometry_item.GetPointIds().SetNumberOfIds(len(topology))

        # Set the connectivity
        for i_local, i_global in enumerate(topology):
            geometry_item.GetPointIds().SetId(i_local, i_global)

        # Add to global cells
        self.grid.InsertNextCell(
            geometry_item.GetCellType(), geometry_item.GetPointIds()
        )

    def _add_data(self, data_container, vtk_geom_type, *, n_new_items=1):
        """Add a data container to the existing global data container of this
        object.

        Args
        ----
        data_container: see self.add_cell
        vtk_geom_type: bme.vtk_geo
            Type of data container that is added
        n_new_items: int
            Number of new items added. This is needed to fill up data fields that are in the
            global data but not in the one that is added.
        """

        # Check if data container already exists. If not, add it and also add
        # previous entries.
        if data_container is not None:
            if vtk_geom_type == VTKGeometry.cell:
                n_items = self.grid.GetNumberOfCells()
            else:
                n_items = self.grid.GetNumberOfPoints()

            for key, item_value in data_container.items():
                # Get the data and the value type (int or float).
                value, data_type = _get_data_value_and_type(item_value)

                # Data type.
                if vtk_geom_type == VTKGeometry.cell:
                    vtk_tensor_type = self._get_vtk_data_type(value)
                else:
                    for item in value:
                        vtk_tensor_type = self._get_vtk_data_type(item)

                # Check if key already exists.
                if key not in self.data[vtk_geom_type, vtk_tensor_type].keys():
                    # Set up the VTK data array.
                    if data_type is VTKType.float:
                        data = _vtk.vtkDoubleArray()
                    else:
                        data = _vtk.vtkIntArray()
                    data.SetName(key)
                    if vtk_tensor_type == VTKTensor.scalar:
                        data.SetNumberOfComponents(1)
                    else:
                        data.SetNumberOfComponents(3)

                    # Add the empty values for all previous cells / points.

                    for i in range(n_items):
                        self._add_single_data_item(data, vtk_tensor_type)
                    self.data[vtk_geom_type, vtk_tensor_type][key] = data

                else:
                    # In this case we just check that the already existing
                    # data has the same type.
                    data_array = self.data[vtk_geom_type, vtk_tensor_type][key]
                    if not _get_vtk_array_type(data_array) == data_type:
                        raise ValueError(
                            (
                                'The existing data with the key "{}"'
                                + ' is of type "{}", but the type you tried to add'
                                + ' is "{}"!'
                            ).format(key, data_array.GetDataTypeAsString(), data_type)
                        )

        # Add to global data. Check if there is something to be added. If not an empty value
        # is added.
        for key_tensor in VTKTensor:
            global_data = self.data[vtk_geom_type, key_tensor]
            if data_container is None:
                data_container = {}

            for key, value in global_data.items():
                # Check if an existing field is also given for this function.
                if key in data_container.keys():
                    # Get the data and the value type (int or float).
                    data_values, _ = _get_data_value_and_type(data_container[key])

                    # Add the given data.
                    if vtk_geom_type == VTKGeometry.cell:
                        self._add_single_data_item(
                            value, key_tensor, non_zero_data=data_values
                        )
                    else:
                        for item in data_values:
                            self._add_single_data_item(
                                value, key_tensor, non_zero_data=item
                            )
                else:
                    # Add empty data.
                    if vtk_geom_type == VTKGeometry.cell:
                        self._add_single_data_item(value, key_tensor)
                    else:
                        for item in range(n_new_items):
                            self._add_single_data_item(value, key_tensor)

    @staticmethod
    def _get_vtk_data_type(data):
        """Return the type of data.

        Check if data matches an expected case.
        """

        if isinstance(data, (list, _np.ndarray)):
            if len(data) == 3:
                return VTKTensor.vector
            raise IndexError(
                f"Only 3d vectors are implemented yet! Got len(data) = {len(data)}"
            )
        elif isinstance(data, _numbers.Number):
            return VTKTensor.scalar

        raise ValueError(f"Data {data} did not match any expected case!")

    @staticmethod
    def _add_single_data_item(data, vtk_tensor_type, non_zero_data=None):
        """Add data to a VTK data array."""

        if _get_vtk_array_type(data) == VTKType.int:
            nan_value = VTK_NAN_INT
        elif _get_vtk_array_type(data) == VTKType.float:
            nan_value = VTK_NAN_FLOAT

        if vtk_tensor_type == VTKTensor.scalar:
            if non_zero_data is None:
                data.InsertNextTuple1(nan_value)
            else:
                data.InsertNextTuple1(non_zero_data)
        else:
            if non_zero_data is None:
                data.InsertNextTuple3(nan_value, nan_value, nan_value)
            else:
                data.InsertNextTuple3(
                    non_zero_data[0], non_zero_data[1], non_zero_data[2]
                )

    def complete_data(self):
        """Add the stored data to the vtk grid."""
        for (key_geom, _key_data), value in self.data.items():
            for vtk_data in value.values():
                if key_geom == VTKGeometry.cell:
                    self.grid.GetCellData().AddArray(vtk_data)
                else:
                    self.grid.GetPointData().AddArray(vtk_data)

    def write_vtk(self, filepath, *, binary=True):
        """Write the VTK geometry and data to a file.

        Args
        ----
        filepath: str
            Path to output file. The file extension should be vtu.
        binary: bool
            If the data should be written encoded in binary or in human readable text.
        """

        # Check if directory for file exits.
        file_directory = _os.path.dirname(filepath)
        if not _os.path.isdir(file_directory):
            raise ValueError(f"Directory {file_directory} does not exist!".format())

        # Initialize VTK writer.
        writer = _vtk.vtkXMLUnstructuredGridWriter()

        # Set the ascii flag.
        if not binary:
            writer.SetDataModeToAscii()

        # Check the file extension.
        _filename, file_extension = _os.path.splitext(filepath)
        if not file_extension.lower() == ".vtu":
            _warnings.warn(f'The extension should be "vtu", got {file_extension}!')

        # Write geometry and data to file.
        writer.SetFileName(filepath)
        writer.SetInputData(self.grid)
        writer.Write()



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Test the performance of the geometric search algorithms."""

import numpy as np
import pytest

from beamme.geometric_search.find_close_points import (
    FindClosePointAlgorithm,
    find_close_points,
)


@pytest.mark.performance
def test_performance_geometric_search_find_close_points_brute_force_cython(
    evaluate_execution_time,
):
    """Test the performance of finding close points using brute force Cython
    algorithm."""

    def repeat_find_random_close_points(n_points, n_runs, algorithm):
        """Repeat finding close points with random points."""
        np.random.seed(seed=1)
        points = np.random.rand(n_points, 3)

        for _ in range(n_runs):
            find_close_points(points, algorithm=algorithm)

    evaluate_execution_time(
        "BeamMe: Find close points (brute force Cython)",
        repeat_find_random_close_points,
        kwargs={
            "n_points": 100,
            "n_runs": 1000,
            "algorithm": FindClosePointAlgorithm.brute_force_cython,
        },
        expected_time=0.025,
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Unit tests for the generic beam mesh creation function utils functions."""

import pytest

from beamme.core.element_beam import Beam3
from beamme.core.mesh import Mesh
from beamme.core.node import NodeCosserat
from beamme.core.rotation import Rotation
from beamme.mesh_creation_functions.beam_line import create_beam_mesh_line


def test_beamme_mesh_creation_functions_beam_generic_start_end_node_error(
    get_default_test_beam_material,
):
    """Check that an error is raised if wrong start and end nodes are given to
    a mesh creation function."""

    # Create mesh object.
    mesh = Mesh()
    mat = get_default_test_beam_material(material_type="base")
    mesh.add(mat)

    # Try to create a line with a starting node that is not in the mesh.
    node = NodeCosserat([0, 0, 0], Rotation())
    args = [mesh, Beam3, mat, [0, 0, 0], [1, 0, 0]]
    kwargs = {"start_node": node}
    with pytest.raises(ValueError):
        create_beam_mesh_line(*args, **kwargs)
    node.coordinates = [1, 0, 0]
    kwargs = {"end_node": node}
    with pytest.raises(ValueError):
        create_beam_mesh_line(*args, **kwargs)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script checks if all Python imports within provided files align with
our coding style, as described in the section "Coding guidelines" in the
repository README.md."""

import ast as _ast
import sys as _sys
from dataclasses import dataclass as _dataclass
from pathlib import Path as _Path
from typing import Optional as _Optional


@_dataclass
class Error:
    """Class to store import errors."""

    filename: str
    message: str
    line_nr: _Optional[int] = None
    column_nr: _Optional[int] = None


class ImportChecker(_ast.NodeVisitor):
    """Class to check if Python imports align with our coding style."""

    def __init__(self, filename: str) -> None:
        """Initialize the ImportChecker.

        Args:
            filename: The filename of the current file (to store for errors)
        """

        self.errors: list[Error] = []
        self.filename = filename

    def visit_Import(self, node: _ast.Import) -> None:
        """Visit an import statement and check if the import aligns with our
        coding style.

        Args:
            node: The import node to check.
        """

        for alias in node.names:
            if alias.asname is None:
                self.errors.append(
                    Error(
                        self.filename,
                        f"Import '{alias.name}' has no alias!",
                        node.lineno,
                        node.col_offset,
                    )
                )
            elif not alias.asname.startswith("_"):
                self.errors.append(
                    Error(
                        self.filename,
                        f"Import '{alias.name}' has alias '{alias.asname}' that doesn't start with '_'!",
                        node.lineno,
                        node.col_offset,
                    )
                )
        self.generic_visit(node)

    def visit_ImportFrom(self, node: _ast.ImportFrom) -> None:
        """Visit an import from statement and check if the import aligns with
        our coding style.

        Args:
            node: The import from node to check.
        """

        for alias in node.names:
            # Check for wildcard imports
            if alias.name == "*":
                self.errors.append(
                    Error(
                        self.filename,
                        "Wildcard imports are not allowed!",
                        node.lineno,
                        node.col_offset,
                    )
                )
                continue

            # Check for private imports from third party libraries
            if not (
                node.module is not None and node.module.startswith("beamme.")
            ) and alias.name.startswith("_"):
                from_module = node.module or f"relative import (level {node.level})"
                self.errors.append(
                    Error(
                        self.filename,
                        f"Import of private functionality '{alias.name}' from '{from_module}' is not allowed!",
                        node.lineno,
                        node.col_offset,
                    )
                )
                continue

            # Check for missing aliases
            if alias.asname is None:
                from_module = node.module or f"relative import (level {node.level})"
                self.errors.append(
                    Error(
                        self.filename,
                        f"Import '{alias.name}' from '{from_module}' has no alias!",
                        node.lineno,
                        node.col_offset,
                    )
                )
            else:
                # Check for correct aliases (internal imports cannot be renamed)
                if node.module is not None and node.module.startswith("beamme."):
                    expected_asname = "_" + alias.name
                    if alias.asname != expected_asname:
                        from_module = (
                            node.module or f"relative import (level {node.level})"
                        )
                        self.errors.append(
                            Error(
                                self.filename,
                                f"Internal import '{alias.name}' from '{from_module}' has incorrect alias '{alias.asname}'! It should be '{expected_asname}'!",
                                node.lineno,
                                node.col_offset,
                            )
                        )
                # Check for correct aliases (external imports can be renamed)
                elif not alias.asname.startswith("_"):
                    from_module = node.module or f"relative import (level {node.level})"
                    self.errors.append(
                        Error(
                            self.filename,
                            f"External import '{alias.name}' from '{from_module}' has incorrect alias '{alias.asname}'! It should be '_{alias.asname}'!",
                            node.lineno,
                            node.col_offset,
                        )
                    )
        self.generic_visit(node)


def check_file(filename: str) -> list[Error]:
    """Check a Python file for import errors.

    Args:
        filename: The filename of the Python file to check.

    Returns:
        List of errors found in the Python file.
    """

    with open(filename, "r") as file:
        content = file.read()
    tree = _ast.parse(content, filename=filename)

    checker = ImportChecker(filename)
    checker.visit(tree)
    return checker.errors


def main() -> None:
    """Check all provided Python files for import errors."""

    errors: list[Error] = []

    for filename in _sys.argv[1:]:
        if not _Path(filename).exists():
            errors.append(Error(filename, "File not found!"))
            continue

        errors.extend(check_file(filename))

    if errors:
        print("Found imports which do not align with our coding style:\n")

        for error in errors:
            if error.line_nr is not None and error.column_nr is not None:
                print(
                    f"    * {error.filename}:{error.line_nr}:{error.column_nr}: {error.message}\n"
                )
            else:
                print(f"{error.filename}: {error.message}\n")

        print(
            "For more information on how to fix these issues, please refer to the contributing guidelines in our README.md."
        )
        _sys.exit(1)
    else:
        _sys.exit(0)


if __name__ == "__main__":
    """Run the main function if the script is executed."""
    main()



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module implements utility functions to create 4C space time
function."""

from typing import List as _List

import numpy as _np

from beamme.core.function import Function as _Function


def create_linear_interpolation_dict(
    times: _List[float], values: _List[float], *, variable_name="var", variable_index=0
):
    """Create a string that describes a variable that is linear interpolated
    over time.

    Args
        times, values:
            Time and values that will be interpolated with piecewise linear functions
        variable_name:
            Name of the created variable
        variable_index:
            Index of this created variable
    """

    if not len(times) == len(values):
        raise ValueError(
            f"The dimensions of time ({len(times)}) and values ({len(values)}) do not match"
        )

    times_appended = _np.array(times)
    values_appended = _np.array(values)
    t_max = _np.max(times_appended)
    times_appended = _np.insert(times_appended, 0, -1000.0, axis=0)
    times_appended = _np.append(times_appended, [t_max + 1000.0])
    values_appended = _np.insert(values_appended, 0, values[0], axis=0)
    values_appended = _np.append(values_appended, values_appended[-1])
    return {
        "VARIABLE": variable_index,
        "NAME": variable_name,
        "TYPE": "linearinterpolation",
        "NUMPOINTS": len(times_appended),
        "TIMES": times_appended.tolist(),
        "VALUES": values_appended.tolist(),
    }


def create_linear_interpolation_function(
    times: _List[float],
    values: _List[float],
    *,
    function_type="SYMBOLIC_FUNCTION_OF_SPACE_TIME",
):
    """Create a function that describes a linear interpolation between the
    given time points and values. Before and after it will be constant.

    Args
    ----
        times, values:
            Time and values that will be interpolated with piecewise linear functions
    """

    function_dict = create_linear_interpolation_dict(times, values, variable_name="var")
    return _Function([{function_type: "var"}, function_dict])


def ensure_length_of_function_array(function_array: _List, length: int = 3):
    """Performs size check of a function array and appends the function array
    to the given length, if a list with only one item is provided.

    Args:
        function_array: list with functions
        length: expected length of function array

    Returns:
        function_array: list with functions with provided length
    """

    # extend items of function automatically if it is only provided once
    if len(function_array) == 1:
        function_array = function_array * length

    if len(function_array) != length:
        raise ValueError(
            f"The function array must have length {length} not {len(function_array)}."
        )
    return function_array



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file contains a function to add the beam interaction conditions for
4C."""

from typing import Optional as _Optional

import beamme.core.conf as _conf
from beamme.core.boundary_condition import BoundaryCondition as _BoundaryCondition
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.core.mesh import Mesh as _Mesh


def get_next_possible_id_for_boundary_condition(
    mesh: _Mesh,
    bc_type: _conf.BoundaryCondition,
    geometry_type: _conf.Geometry,
    condition_string: str,
) -> int:
    """Returns the next possible id, which can be used for a boundary condition
    based on all previous added boundary conditions within a mesh.

    It returns the first ID which is not yet occupied within the existing boundary conditions
    w.r.t. to the given search_string and regex group index.

    Args:
        mesh: Mesh containing already set boundary conditions.
        bc_type: Type of the boundary condition to be searched.
        geometry_type: Geometry type of the boundary condition.
        condition_string: String defining the condition ID tag.

    Returns:
        id: Smallest available ID
    """

    found_conditions = []

    # loop through every possible geometry and find the conditions
    if (bc_type, geometry_type) in mesh.boundary_conditions:
        for bc_condition in mesh.boundary_conditions[(bc_type, geometry_type)]:
            found_conditions.append(bc_condition)

    if not found_conditions:
        # return starting id, since no conditions of type has been added.
        return 0
    else:
        existing_ids = []

        # compare string of each condition with input and store existing ids
        for bc in found_conditions:
            if condition_string in bc.data:
                existing_ids.append(bc.data[condition_string])
            else:
                raise KeyError(
                    f"The key {condition_string} is not in the data {bc.data}"
                )

        # return lowest found id
        return min(set(range(len(existing_ids) + 1)) - set(existing_ids))


def add_beam_interaction_condition(
    mesh: _Mesh,
    geometry_set_1: _GeometrySet,
    geometry_set_2: _GeometrySet,
    bc_type: _conf.BoundaryCondition,
    *,
    id: _Optional[int] = None,
) -> int:
    """Adds a pair of beam interaction boundary conditions to the given mesh
    and estimates automatically the id of them based on all previously added
    boundary conditions of the mesh.

    Args:
        mesh: Mesh to which the boundary conditions will be added.
        geometry_set_1: GeometrySet 1 for beam interaction boundary condition
        geometry_set_2: GeometrySet 2 for beam interaction boundary condition
        id: id of the two conditions

    Returns:
        id: Used id for the created condition.
    """

    condition_string = "COUPLING_ID"
    if id is None:
        id = get_next_possible_id_for_boundary_condition(
            mesh,
            bc_type,
            geometry_set_1.geometry_type,
            condition_string=condition_string,
        )

        id_2 = get_next_possible_id_for_boundary_condition(
            mesh,
            bc_type,
            geometry_set_2.geometry_type,
            condition_string=condition_string,
        )

        if not id == id_2:
            raise ValueError(
                f"The estimated IDs {id} and {id_2} do not match. Check Inputfile."
            )

    # Creates the two conditions with the same ID.
    for geometry_set in [geometry_set_1, geometry_set_2]:
        mesh.add(
            _BoundaryCondition(
                geometry_set, data={condition_string: id}, bc_type=bc_type
            )
        )

    return id



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file defines functions to dump mesh items for 4C."""

from typing import Any as _Any

from beamme.core.boundary_condition import BoundaryCondition as _BoundaryCondition
from beamme.core.conf import bme as _bme
from beamme.core.coupling import Coupling as _Coupling
from beamme.core.element_volume import VolumeElement as _VolumeElement
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.core.geometry_set import GeometrySetNodes as _GeometrySetNodes
from beamme.core.node import ControlPoint as _ControlPoint
from beamme.core.node import Node as _Node
from beamme.core.nurbs_patch import NURBSPatch as _NURBSPatch
from beamme.four_c.four_c_types import (
    BeamKirchhoffParametrizationType as _BeamKirchhoffParametrizationType,
)
from beamme.four_c.four_c_types import BeamType as _BeamType
from beamme.four_c.input_file_mappings import (
    INPUT_FILE_MAPPINGS as _INPUT_FILE_MAPPINGS,
)


def dump_node(node):
    """Return the representation of a node in the 4C input file."""

    if isinstance(node, _ControlPoint):
        return {
            "id": node.i_global + 1,
            "COORD": node.coordinates,
            "data": {"type": "CP", "weight": node.weight},
        }
    elif isinstance(node, _Node):
        return {
            "id": node.i_global + 1,
            "COORD": node.coordinates,
            "data": {"type": "NODE"},
        }
    else:
        raise TypeError(f"Got unexpected item of type {type(node)}")


def dump_solid_element(solid_element):
    """Return a dict with the items representing the given solid element."""

    return {
        "id": solid_element.i_global + 1,
        "cell": {
            "type": _INPUT_FILE_MAPPINGS["element_type_to_four_c_string"][
                type(solid_element)
            ],
            "connectivity": solid_element.nodes,
        },
        "data": solid_element.data,
    }


def dump_coupling(coupling):
    """Return the input file representation of the coupling condition."""

    if isinstance(coupling.data, dict):
        data = coupling.data
    else:
        # In this case we have to check which beams are connected to the node.
        # TODO: Coupling also makes sense for different beam types, this can
        # be implemented at some point.
        nodes = coupling.geometry_set.get_points()
        connected_elements = [
            element for node in nodes for element in node.element_link
        ]
        element_types = {type(element) for element in connected_elements}
        if len(element_types) > 1:
            raise TypeError(
                f"Expected a single connected type of beam elements, got {element_types}"
            )
        element_type = element_types.pop()
        if element_type.four_c_beam_type is _BeamType.kirchhoff:
            unique_parametrization_flags = {
                _BeamKirchhoffParametrizationType[
                    type(element).four_c_element_data["PARAMETRIZATION"]
                ]
                for element in connected_elements
            }
            if (
                len(unique_parametrization_flags) > 1
                or not unique_parametrization_flags.pop()
                == _BeamKirchhoffParametrizationType.rot
            ):
                raise TypeError(
                    "Couplings for Kirchhoff beams and tangent "
                    "based parametrization not yet implemented."
                )

        data = element_type.get_coupling_dict(coupling.data)

    return {"E": coupling.geometry_set.i_global + 1, **data}


def dump_geometry_set(geometry_set):
    """Return a list with the data describing this set."""

    # Sort nodes based on their global index
    nodes = sorted(geometry_set.get_all_nodes(), key=lambda n: n.i_global)

    if not nodes:
        raise ValueError("Writing empty geometry sets is not supported")

    return [
        {
            "type": "NODE",
            "node_id": node.i_global + 1,
            "d_type": _INPUT_FILE_MAPPINGS["geometry_sets_geometry_to_entry_name"][
                geometry_set.geometry_type
            ],
            "d_id": geometry_set.i_global + 1,
        }
        for node in nodes
    ]


def dump_nurbs_patch_knotvectors(input_file, nurbs_patch) -> None:
    """Set the knot vectors of the NURBS patch in the input file."""

    patch_data: dict[str, _Any] = {
        "KNOT_VECTORS": [],
    }

    for dir_manifold in range(nurbs_patch.get_nurbs_dimension()):
        knotvector = nurbs_patch.knot_vectors[dir_manifold]
        num_knots = len(knotvector)

        # Check the type of knot vector, in case that the multiplicity of the first and last
        # knot vectors is not p + 1, then it is a closed (periodic) knot vector, otherwise it
        # is an open (interpolated) knot vector.
        knotvector_type = "Interpolated"

        for i in range(nurbs_patch.polynomial_orders[dir_manifold] - 1):
            if (abs(knotvector[i] - knotvector[i + 1]) > _bme.eps_knot_vector) or (
                abs(knotvector[num_knots - 2 - i] - knotvector[num_knots - 1 - i])
                > _bme.eps_knot_vector
            ):
                knotvector_type = "Periodic"
                break

        patch_data["KNOT_VECTORS"].append(
            {
                "DEGREE": nurbs_patch.polynomial_orders[dir_manifold],
                "TYPE": knotvector_type,
                "KNOTS": [
                    knot_vector_val
                    for knot_vector_val in nurbs_patch.knot_vectors[dir_manifold]
                ],
            }
        )

    if "STRUCTURE KNOTVECTORS" in input_file:
        # Get all existing patches in the input file - they will be added to the
        # input file again at the end of this function. By doing it this way, the
        # FourCIPP type converter will be applied to the current patch.
        # This also means that we apply the type converter again already existing
        # patches. But, with the usual number of patches and data size, this
        # should not lead to a measurable performance impact.
        patches = input_file.pop("STRUCTURE KNOTVECTORS")["PATCHES"]
    else:
        patches = []

    patch_data["ID"] = nurbs_patch.i_nurbs_patch + 1
    patches.append(patch_data)
    input_file.add({"STRUCTURE KNOTVECTORS": {"PATCHES": patches}})


def dump_nurbs_patch_elements(nurbs_patch: _NURBSPatch) -> list[dict[str, _Any]]:
    """Return a list with all the element definitions contained in this
    patch."""

    if nurbs_patch.i_global is None:
        raise ValueError(
            "i_global is not set, make sure that the NURBS patch is added to the mesh"
        )

    # Check the material
    nurbs_patch._check_material()

    patch_elements = []
    j = 0

    for knot_span in nurbs_patch.get_knot_span_iterator():
        element_cps_ids = nurbs_patch.get_ids_ctrlpts(*knot_span)
        connectivity = [nurbs_patch.nodes[i] for i in element_cps_ids]
        num_cp = len(connectivity)

        patch_elements.append(
            {
                "id": nurbs_patch.i_global + j + 1,
                "cell": {
                    "type": f"NURBS{num_cp}",
                    "connectivity": connectivity,
                },
                "data": {
                    "type": _INPUT_FILE_MAPPINGS["nurbs_type_to_default_four_c_type"][
                        type(nurbs_patch)
                    ],
                    "MAT": nurbs_patch.material,
                    **(nurbs_patch.data if nurbs_patch.data else {}),
                },
            }
        )
        j += 1

    return patch_elements


def dump_item_to_list(dumped_list, item) -> None:
    """General function to dump items to a 4C input file."""
    if hasattr(item, "dump_to_list"):
        dumped_list.append(item.dump_to_list())
    elif isinstance(item, _Node):
        dumped_list.append(dump_node(item))
    elif isinstance(item, _VolumeElement):
        dumped_list.append(dump_solid_element(item))
    elif isinstance(item, _GeometrySet) or isinstance(item, _GeometrySetNodes):
        dumped_list.extend(dump_geometry_set(item))
    elif isinstance(item, _NURBSPatch):
        dumped_list.extend(dump_nurbs_patch_elements(item))
    elif isinstance(item, _BoundaryCondition):
        if item.geometry_set.i_global is None:
            raise ValueError("i_global is not set")
        dumped_list.append(
            {
                "E": item.geometry_set.i_global + 1,
                **item.data,
            }
        )
    elif isinstance(item, _Coupling):
        dumped_list.append(dump_coupling(item))
    else:
        raise TypeError(f"Could not dump {item}")


def dump_item_to_section(input_file, item) -> None:
    """This function dumps information of mesh items to general input file
    sections, e.g., knotvectors for NURBS."""
    if isinstance(item, _NURBSPatch):
        dump_nurbs_patch_knotvectors(input_file, item)



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This script is used to test general core functionality of BeamMe."""

import autograd.numpy as npAD
import numpy as np

from beamme.core.geometry_set import GeometryName, GeometrySet
from beamme.core.mesh import Mesh
from beamme.core.node import NodeCosserat
from beamme.core.rotation import Rotation
from beamme.four_c.element_beam import Beam3rHerm2Line3
from beamme.mesh_creation_functions.beam_arc import (
    create_beam_mesh_arc_segment_via_rotation,
)
from beamme.mesh_creation_functions.beam_parametric_curve import (
    create_beam_mesh_parametric_curve,
)


def test_integration_core_close_beam(
    get_default_test_beam_material,
    assert_results_close,
    get_corresponding_reference_file_path,
):
    """Create a circle with different methods.
    - Create the mesh manually by creating the nodes and connecting them to
        the elements.
    - Create one full circle and connect it to its beginning.
    - Create two half circle and connect their start / end nodes.
    All of those methods should give the exact same mesh.
    Both variants are also tried with different rotations at the beginning.
    """

    # Parameters for this test case.
    n_el = 3
    R = 1.235
    additional_rotation = Rotation([0, 1, 0], 0.5)

    # Define material.
    mat = get_default_test_beam_material(material_type="reissner")

    def create_mesh_manually(start_rotation):
        """Create the full circle manually."""
        mesh = Mesh()
        mesh.add(mat)

        # Add nodes.
        for i in range(4 * n_el):
            basis = start_rotation * Rotation([0, 0, 1], np.pi * 0.5)
            r = [R, 0, 0]
            node = NodeCosserat(r, basis)
            rotation = Rotation([0, 0, 1], 0.5 * i * np.pi / n_el)
            node.rotate(rotation, origin=[0, 0, 0])
            mesh.nodes.append(node)

        # Add elements.
        for i in range(2 * n_el):
            node_index = [2 * i, 2 * i + 1, 2 * i + 2]
            nodes = []
            for index in node_index:
                if index == len(mesh.nodes):
                    nodes.append(mesh.nodes[0])
                else:
                    nodes.append(mesh.nodes[index])
            element = Beam3rHerm2Line3(mat, nodes)
            mesh.add(element)

        # Add sets.
        geom_set = GeometryName()
        geom_set["start"] = GeometrySet(mesh.nodes[0])
        geom_set["end"] = GeometrySet(mesh.nodes[0])
        geom_set["line"] = GeometrySet(mesh.elements)
        mesh.add(geom_set)
        return mesh

    def one_full_circle_closed(function, argument_list, additional_rotation=None):
        """Create one full circle and connect it to itself."""

        mesh = Mesh()

        if additional_rotation is not None:
            start_rotation = additional_rotation * Rotation([0, 0, 1], np.pi * 0.5)
            mesh.add(NodeCosserat([R, 0, 0], start_rotation))
            beam_sets = function(
                mesh,
                start_node=mesh.nodes[0],
                close_beam=True,
                **(argument_list),
            )
        else:
            beam_sets = function(mesh, close_beam=True, **(argument_list))
        mesh.add(beam_sets)
        return mesh

    def two_half_circles_closed(function, argument_list, additional_rotation=None):
        """Create two half circles and close them, by reusing the connecting
        nodes."""

        mesh = Mesh()

        if additional_rotation is not None:
            start_rotation = additional_rotation * Rotation([0, 0, 1], np.pi * 0.5)
            mesh.add(NodeCosserat([R, 0, 0], start_rotation))
            set_1 = function(mesh, start_node=mesh.nodes[0], **(argument_list[0]))
        else:
            set_1 = function(mesh, **(argument_list[0]))

        set_2 = function(
            mesh,
            start_node=set_1["end"],
            end_node=set_1["start"],
            **(argument_list[1]),
        )

        # Add sets.
        geom_set = GeometryName()
        geom_set["start"] = GeometrySet(set_1["start"])
        geom_set["end"] = GeometrySet(set_2["end"])
        geom_set["line"] = GeometrySet([set_1["line"], set_2["line"]])
        mesh.add(geom_set)

        return mesh

    def get_arguments_arc_segment(circle_type):
        """Return the arguments for the arc segment function."""
        if circle_type == 0:
            # Full circle.
            arg_rot_angle = np.pi / 2
            arg_angle = 2 * np.pi
            arg_n_el = 2 * n_el
        elif circle_type == 1:
            # First half circle.
            arg_rot_angle = np.pi / 2
            arg_angle = np.pi
            arg_n_el = n_el
        elif circle_type == 2:
            # Second half circle.
            arg_rot_angle = 3 * np.pi / 2
            arg_angle = np.pi
            arg_n_el = n_el
        return {
            "beam_class": Beam3rHerm2Line3,
            "material": mat,
            "center": [0, 0, 0],
            "axis_rotation": Rotation([0, 0, 1], arg_rot_angle),
            "radius": R,
            "angle": arg_angle,
            "n_el": arg_n_el,
        }

    def circle_function(t):
        """Function for the circle."""
        return R * npAD.array([npAD.cos(t), npAD.sin(t)])

    def get_arguments_curve(circle_type):
        """Return the arguments for the curve function."""
        if circle_type == 0:
            # Full circle.
            arg_interval = [0, 2 * np.pi]
            arg_n_el = 2 * n_el
        elif circle_type == 1:
            # First half circle.
            arg_interval = [0, np.pi]
            arg_n_el = n_el
        elif circle_type == 2:
            # Second half circle.
            arg_interval = [np.pi, 2 * np.pi]
            arg_n_el = n_el
        return {
            "beam_class": Beam3rHerm2Line3,
            "material": mat,
            "function": circle_function,
            "interval": arg_interval,
            "n_el": arg_n_el,
        }

    # Check the meshes without additional rotation.
    assert_results_close(
        get_corresponding_reference_file_path(), create_mesh_manually(Rotation())
    )
    assert_results_close(
        get_corresponding_reference_file_path(),
        one_full_circle_closed(
            create_beam_mesh_arc_segment_via_rotation, get_arguments_arc_segment(0)
        ),
    )
    assert_results_close(
        get_corresponding_reference_file_path(),
        two_half_circles_closed(
            create_beam_mesh_arc_segment_via_rotation,
            [get_arguments_arc_segment(1), get_arguments_arc_segment(2)],
        ),
    )
    assert_results_close(
        get_corresponding_reference_file_path(),
        one_full_circle_closed(
            create_beam_mesh_parametric_curve, get_arguments_curve(0)
        ),
    )
    assert_results_close(
        get_corresponding_reference_file_path(),
        two_half_circles_closed(
            create_beam_mesh_parametric_curve,
            [get_arguments_curve(1), get_arguments_curve(2)],
        ),
    )

    # Check the meshes with additional rotation.
    additional_identifier = "rotation"
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=additional_identifier
        ),
        create_mesh_manually(additional_rotation),
    )
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=additional_identifier
        ),
        one_full_circle_closed(
            create_beam_mesh_arc_segment_via_rotation,
            get_arguments_arc_segment(0),
            additional_rotation=additional_rotation,
        ),
    )
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=additional_identifier
        ),
        two_half_circles_closed(
            create_beam_mesh_arc_segment_via_rotation,
            [get_arguments_arc_segment(1), get_arguments_arc_segment(2)],
            additional_rotation=additional_rotation,
        ),
    )
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=additional_identifier
        ),
        one_full_circle_closed(
            create_beam_mesh_parametric_curve,
            get_arguments_curve(0),
            additional_rotation=additional_rotation,
        ),
    )
    assert_results_close(
        get_corresponding_reference_file_path(
            additional_identifier=additional_identifier
        ),
        two_half_circles_closed(
            create_beam_mesh_parametric_curve,
            [get_arguments_curve(1), get_arguments_curve(2)],
            additional_rotation=additional_rotation,
        ),
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
r"""This module defines geometric search functionality.

This module contains functionality to find points close to each other in
a point cloud. Currently, three different implementations for the actual
search algorithm are available (depending on your installation/setup):

- `brute_force_cython`: A brute force algorithm implemented in Cython,
  scales with $\mathcal{O}(n^2)$, but for a small number of points
  ($n<200$) this is the fastest algorithm since compared to the others
  it does not have any setup costs.

- `kd_tree_scipy`: Uses a
  [bounding volume hierarchy (BVH)](https://en.wikipedia.org/wiki/Bounding_volume_hierarchy)
  implementation provided by
  [scipy](https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.KDTree.html).
  This scales with $\mathcal{O}(n\ \log{n})$.

- `boundary_volume_hierarchy_arborx` : Uses a
  [bounding volume hierarchy (BVH)](https://en.wikipedia.org/wiki/Bounding_volume_hierarchy)
  implementation provided by [ArborX](https://github.com/arborx/ArborX).
  This also scales with $\mathcal{O}(n\ \log{n})$
  but due to a more optimised implementation is a few times faster
  than the scipy implementation.

The `find_close_points` function automatically chooses the fastest
(available) implementation for the given point array.

Consult the `README.md` regarding install and testing options for
different implementations.
"""



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module implements a class to represent boundary conditions."""

import warnings as _warnings
from typing import Dict as _Dict
from typing import Optional as _Optional
from typing import Union as _Union

import beamme.core.conf as _conf
from beamme.core.base_mesh_item import BaseMeshItem as _BaseMeshItem
from beamme.core.conf import bme as _bme
from beamme.core.container import ContainerBase as _ContainerBase
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.core.geometry_set import GeometrySetBase as _GeometrySetBase
from beamme.utils.nodes import find_close_nodes as _find_close_nodes


class BoundaryConditionBase(_BaseMeshItem):
    """Base class for boundary conditions."""

    def __init__(
        self,
        geometry_set: _GeometrySetBase,
        bc_type: _Union[_conf.BoundaryCondition, str],
        **kwargs,
    ):
        """Initialize the boundary condition.

        Args:
            geometry_set: Geometry that this boundary condition acts on.
            bc_type: Type of the boundary condition.
        """

        super().__init__(**kwargs)
        self.bc_type = bc_type
        self.geometry_set = geometry_set


class BoundaryCondition(BoundaryConditionBase):
    """This object represents one boundary condition, e.g., Dirichlet, Neumann,
    ..."""

    def __init__(
        self,
        geometry_set: _GeometrySetBase,
        data: _Dict,
        bc_type: _Union[_conf.BoundaryCondition, str],
        *,
        double_nodes: _Optional[_conf.DoubleNodes] = None,
        **kwargs,
    ):
        """Initialize the object.

        Args:
            geometry_set: Geometry that this boundary condition acts on.
            data: Data defining the properties of this boundary condition.
            bc_type: If this is a string, this will be the section that
                this BC will be added to. If it is a bme.bc, the section will
                be determined automatically.
            double_nodes: Depending on this parameter, it will be checked if point
                Neumann conditions do contain nodes at the same spatial positions.
        """

        super().__init__(geometry_set, bc_type, data=data, **kwargs)
        self.double_nodes = double_nodes

        # Perform some sanity checks for this boundary condition.
        self.check()

    def check(self):
        """Check for point Neumann boundaries that there is not a double Node
        in the set.

        Duplicate nodes in a point Neumann boundary condition can lead
        to the same force being applied multiple times at the same
        spatial position, which results in incorrect load application.
        """

        if self.double_nodes is _bme.double_nodes.keep:
            return

        if (
            self.bc_type == _bme.bc.neumann
            and self.geometry_set.geometry_type == _bme.geo.point
        ):
            my_nodes = self.geometry_set.get_points()
            partners = _find_close_nodes(my_nodes)
            # Create a list with nodes that will not be kept in the set.
            double_node_list = []
            for node_list in partners:
                for i, node in enumerate(node_list):
                    if i > 0:
                        double_node_list.append(node)
            if (
                len(double_node_list) > 0
                and self.double_nodes is _bme.double_nodes.remove
            ):
                # Create the a new geometry set with the unique nodes.
                self.geometry_set = _GeometrySet(
                    [node for node in my_nodes if (node not in double_node_list)]
                )
            elif len(double_node_list) > 0:
                _warnings.warn(
                    "There are overlapping nodes in this point Neumann boundary, and it is not "
                    "specified on how to handle them!"
                )


class BoundaryConditionContainer(_ContainerBase):
    """A class to group boundary conditions together.

    The key of the dictionary are (bc_type, geometry_type).
    """

    def __init__(self, *args, **kwargs):
        """Initialize the container and create the default keys in the map."""
        super().__init__(*args, **kwargs)

        self.item_types = [BoundaryConditionBase]

        for bc_key in _bme.bc:
            for geometry_key in _bme.geo:
                self[(bc_key, geometry_key)] = []



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module implements a class to couple geometry together."""

from typing import List as _List
from typing import Union as _Union

import numpy as _np

import beamme.core.conf as _conf
from beamme.core.boundary_condition import (
    BoundaryConditionBase as _BoundaryConditionBase,
)
from beamme.core.conf import bme as _bme
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.core.geometry_set import GeometrySetBase as _GeometrySetBase
from beamme.core.node import Node as _Node


class Coupling(_BoundaryConditionBase):
    """Represents a coupling between geometries in 4C."""

    def __init__(
        self,
        geometry: _Union[_GeometrySetBase, _List[_Node]],
        coupling_type: _Union[_conf.BoundaryCondition, str],
        coupling_dof_type: _Union[_conf.CouplingDofType, dict],
        *,
        check_overlapping_nodes: bool = True,
    ):
        """Initialize this object.

        Args:
            geometry: Geometry set or nodes that should be coupled.
            coupling_type: If this is a string, this will be the section that
                this coupling will be added to. If it is a bme.bc, the section
                will be determined automatically.
            coupling_dof_type: If this is a dictionary it is the dictionary
                that will be used in the input file, otherwise it has to be
                of type bme.coupling_dof.
            check_overlapping_nodes: If all nodes of this coupling condition
                have to be at the same physical position.
        """

        if isinstance(geometry, _GeometrySetBase):
            pass
        elif isinstance(geometry, list):
            geometry = _GeometrySet(geometry)
        else:
            raise TypeError(
                f"Coupling expects a GeometrySetBase item, got {type(geometry)}"
            )

        # Couplings only work for point sets
        if geometry.geometry_type is not _bme.geo.point:
            raise TypeError("Couplings are only implemented for point sets.")

        super().__init__(geometry, bc_type=coupling_type, data=coupling_dof_type)
        self.check_overlapping_nodes = check_overlapping_nodes

        # Perform sanity checks for this boundary condition
        self.check()

    def check(self):
        """Check that all nodes that are coupled have the same position
        (depending on the check_overlapping_nodes parameter)."""

        if not self.check_overlapping_nodes:
            return

        nodes = self.geometry_set.get_points()
        diff = _np.zeros([len(nodes), 3])
        for i, node in enumerate(nodes):
            # Get the difference to the first node
            diff[i, :] = node.coordinates - nodes[0].coordinates
        if _np.max(_np.linalg.norm(diff, axis=1)) > _bme.eps_pos:
            raise ValueError(
                "The nodes given to Coupling do not have the same position."
            )


def coupling_factory(
    geometry: _Union[_GeometrySetBase, _List[_Node]],
    coupling_type: _conf.BoundaryCondition,
    coupling_dof_type: _Union[_conf.CouplingDofType, dict],
    **kwargs,
) -> list[Coupling]:
    """Create coupling conditions for the nodes in geometry.

    Args:
        geometry: Geometry set or nodes that should be coupled.
        coupling_type: If this is a string, this will be the section that
            this coupling will be added to. If it is a bme.bc, the section
            will be determined automatically.
        coupling_dof_type: If this is a dictionary it is the dictionary
            that will be used in the input file, otherwise it has to be
            of type bme.coupling_dof.
        kwargs: Will be passed to constructor of `Coupling`.

    Returns:
        A list of coupling objects representing the created coupling conditions.
            - By default, a single coupling object is created that couples all nodes in the given geometry.
            - If the selected coupling type requires pairwise coupling (e.g., due to solver restrictions),
              multiple coupling objects are returned, each coupling a pair of nodes accordingly.
    """

    if not coupling_type.is_point_coupling_pairwise():
        return [Coupling(geometry, coupling_type, coupling_dof_type, **kwargs)]
    else:
        if isinstance(geometry, _GeometrySetBase):
            nodes = geometry.get_points()
        else:
            nodes = geometry
        main_node = nodes[0]
        return [
            Coupling([main_node, node], coupling_type, coupling_dof_type, **kwargs)
            for node in nodes[1:]
        ]



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file implements basic classes to manage materials."""

import numpy as _np

from beamme.core.base_mesh_item import BaseMeshItem as _BaseMeshItem


class Material(_BaseMeshItem):
    """Base class for all materials."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __deepcopy__(self, memo):
        """When deepcopy is called on a mesh, we do not want the materials to
        be copied, as this will result in multiple equal materials in the input
        file."""

        # Add this object to the memo dictionary.
        memo[id(self)] = self

        # Return this object again, as no copy should be created.
        return self


class MaterialBeamBase(Material):
    """Base class for all beam materials."""

    def __init__(
        self,
        radius=None,
        material_string=None,
        youngs_modulus=None,
        nu=0.0,
        density=0.0,
        interaction_radius=None,
        **kwargs,
    ):
        """Set the material values that all beams have."""

        super().__init__(**kwargs)

        self.radius = radius
        self.material_string = material_string
        self.youngs_modulus = youngs_modulus
        self.nu = nu
        self.density = density
        self.interaction_radius = interaction_radius
        self.area = None
        self.mom2 = None
        self.mom3 = None
        self.polar = None

    def calc_area_stiffness(self):
        """Calculate the relevant stiffness terms and the area for the given
        beam."""
        area = 4 * self.radius**2 * _np.pi * 0.25
        mom2 = self.radius**4 * _np.pi * 0.25
        mom3 = mom2
        polar = mom2 + mom3
        return area, mom2, mom3, polar


class MaterialSolidBase(Material):
    """Base class for all solid materials."""

    pass



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This function allows to reorder the connectivity of solid shell elements
such that the solid shell direction is correctly represented."""

from typing import List as _List

import numpy as _np
import pyvista as _pv

from beamme.core.element import Element as _Element
from beamme.core.element_volume import VolumeElement as _VolumeElement
from beamme.core.element_volume import VolumeHEX8 as _VolumeHEX8
from beamme.utils.nodes import get_nodal_coordinates as _get_nodal_coordinates


def shape_functions_hex8(xi1, xi2, xi3):
    """Return the shape functions for a hex8 element."""
    shape_functions = _np.zeros(8)
    one_over_eight = 0.125
    shape_functions[0] = one_over_eight * (1 - xi1) * (1 - xi2) * (1 - xi3)
    shape_functions[1] = one_over_eight * (1 + xi1) * (1 - xi2) * (1 - xi3)
    shape_functions[2] = one_over_eight * (1 + xi1) * (1 + xi2) * (1 - xi3)
    shape_functions[3] = one_over_eight * (1 - xi1) * (1 + xi2) * (1 - xi3)
    shape_functions[4] = one_over_eight * (1 - xi1) * (1 - xi2) * (1 + xi3)
    shape_functions[5] = one_over_eight * (1 + xi1) * (1 - xi2) * (1 + xi3)
    shape_functions[6] = one_over_eight * (1 + xi1) * (1 + xi2) * (1 + xi3)
    shape_functions[7] = one_over_eight * (1 - xi1) * (1 + xi2) * (1 + xi3)
    return shape_functions


def shape_functions_derivative_hex8(xi1, xi2, xi3):
    """Return the derivative of the shape functions for a hex8 element."""
    derivatives = _np.zeros((3, 8))
    one_over_eight = 0.125

    # Derivatives with respect to xi1
    derivatives[0, 0] = -one_over_eight * (1 - xi2) * (1 - xi3)
    derivatives[0, 1] = one_over_eight * (1 - xi2) * (1 - xi3)
    derivatives[0, 2] = one_over_eight * (1 + xi2) * (1 - xi3)
    derivatives[0, 3] = -one_over_eight * (1 + xi2) * (1 - xi3)
    derivatives[0, 4] = -one_over_eight * (1 - xi2) * (1 + xi3)
    derivatives[0, 5] = one_over_eight * (1 - xi2) * (1 + xi3)
    derivatives[0, 6] = one_over_eight * (1 + xi2) * (1 + xi3)
    derivatives[0, 7] = -one_over_eight * (1 + xi2) * (1 + xi3)

    # Derivatives with respect to xi2
    derivatives[1, 0] = -one_over_eight * (1 - xi1) * (1 - xi3)
    derivatives[1, 1] = -one_over_eight * (1 + xi1) * (1 - xi3)
    derivatives[1, 2] = one_over_eight * (1 + xi1) * (1 - xi3)
    derivatives[1, 3] = one_over_eight * (1 - xi1) * (1 - xi3)
    derivatives[1, 4] = -one_over_eight * (1 - xi1) * (1 + xi3)
    derivatives[1, 5] = -one_over_eight * (1 + xi1) * (1 + xi3)
    derivatives[1, 6] = one_over_eight * (1 + xi1) * (1 + xi3)
    derivatives[1, 7] = one_over_eight * (1 - xi1) * (1 + xi3)

    # Derivatives with respect to xi3
    derivatives[2, 0] = -one_over_eight * (1 - xi1) * (1 - xi2)
    derivatives[2, 1] = -one_over_eight * (1 + xi1) * (1 - xi2)
    derivatives[2, 2] = -one_over_eight * (1 + xi1) * (1 + xi2)
    derivatives[2, 3] = -one_over_eight * (1 - xi1) * (1 + xi2)
    derivatives[2, 4] = one_over_eight * (1 - xi1) * (1 - xi2)
    derivatives[2, 5] = one_over_eight * (1 + xi1) * (1 - xi2)
    derivatives[2, 6] = one_over_eight * (1 + xi1) * (1 + xi2)
    derivatives[2, 7] = one_over_eight * (1 - xi1) * (1 + xi2)

    return derivatives


def get_hex8_element_center_and_jacobian_mapping(element):
    """Return the center of a hex8 element and the Jacobian mapping for that
    point."""

    nodal_coordinates = _get_nodal_coordinates(element.nodes)
    if not len(nodal_coordinates) == 8:
        raise ValueError(f"Expected 8 nodes, got {len(nodal_coordinates)}")

    N = shape_functions_hex8(0, 0, 0)
    dN = shape_functions_derivative_hex8(0, 0, 0)

    reference_position_center = _np.dot(N, nodal_coordinates)
    jacobian_center = _np.dot(dN, nodal_coordinates)

    return reference_position_center, jacobian_center


def get_reordering_index_thickness(jacobian, *, identify_threshold=None):
    """Return the reordering index from the Jacobian such that the thinnest
    direction is the 3rd parameter direction.

    Additionally it is checked, that the thinnest direction is at least
    identify_threshold times smaller than the next thinnest, to avoid
    wrongly detected directions.
    """

    # The direction with the smallest parameter derivative is the thickness direction
    parameter_derivative_norms = [
        _np.linalg.norm(parameter_direction) for parameter_direction in jacobian
    ]
    thickness_direction = _np.argmin(parameter_derivative_norms)

    if identify_threshold is not None:
        # Check that the minimal parameter direction is at least a given factor
        # smaller than the other directions. This helps to identify cases where
        # it is unlikely that a unique direction is found.
        min_norm = parameter_derivative_norms[thickness_direction]
        relative_difference = [
            parameter_derivative_norms[i] / min_norm
            for i in range(3)
            if not i == thickness_direction
        ]
        if _np.min(relative_difference) < 1.5:
            raise ValueError("Could not uniquely identify the thickness direction.")

    return thickness_direction


def get_reordering_index_director_projection(
    jacobian, director, *, identify_threshold=None
):
    """Return the reordering index from the Jacobian such that the thickness
    direction is the one that has the largest dot product with the given
    director."""

    projections = []
    for parameter_director in jacobian:
        parameter_director = parameter_director / _np.linalg.norm(parameter_director)
        projections.append(_np.abs(_np.dot(director, parameter_director)))
    thickness_direction = _np.argmax(projections)

    if identify_threshold is not None:
        # Check that the maximal dot product is at least a given factor larger than
        # the other projections. This helps to identify cases where it is unlikely
        # that a unique direction is found.
        max_norm = projections[thickness_direction]
        relative_difference = [
            projections[i] / max_norm for i in range(3) if not i == thickness_direction
        ]
        if _np.max(relative_difference) > 1.0 / identify_threshold:
            raise ValueError("Could not uniquely identify the thickness direction.")

    return thickness_direction


def set_solid_shell_thickness_direction(
    elements: _List[_Element],
    *,
    selection_type="thickness",
    director=None,
    director_function=None,
    identify_threshold=2.0,
):
    """Set the solid shell directions for all solid shell elements in the
    element list.

    Args:
    ----
    elements:
        A list containing all elements that should be checked
    selection_type:
        The type of algorithm that shall be used to select the thickness direction
            "thickness":
                The "smallest" dimension of the element will be set to the
                thickness direction
            "projection_director":
                The parameter director that aligns most with a given director
                will be set as the thickness direction
            "projection_director_function":
                The parameter director that aligns most with a director obtained
                by a given function of the element centroid coordinates will be
                set as the thickness direction
    identify_threshold: float/None
        To ensure that the found directions are well-defined, i.e., that not multiple
        directions are almost equally suited to the thickness direction.
    """

    if len(elements) == 0:
        raise ValueError("Expected a non empty element list")

    for element in elements:
        if (
            isinstance(element, _VolumeHEX8)
            and isinstance(element.data, dict)
            and element.data.get("TECH") == "shell_eas_ans"
        ):
            # Get the element center and the Jacobian at the center
            (
                reference_position_center,
                jacobian_center,
            ) = get_hex8_element_center_and_jacobian_mapping(element)

            # Depending on the chosen method, get the thickness direction
            if selection_type == "thickness":
                thickness_direction = get_reordering_index_thickness(
                    jacobian_center, identify_threshold=identify_threshold
                )
            elif selection_type == "projection_director":
                thickness_direction = get_reordering_index_director_projection(
                    jacobian_center, director, identify_threshold=identify_threshold
                )
            elif selection_type == "projection_director_function":
                director = director_function(reference_position_center)
                thickness_direction = get_reordering_index_director_projection(
                    jacobian_center, director, identify_threshold=identify_threshold
                )
            else:
                raise ValueError(
                    f'Got unexpected selection_type of value "{selection_type}"'
                )

            n_apply_mapping = 0
            if thickness_direction == 2:
                # We already have the orientation we want
                continue
            elif thickness_direction == 1:
                # We need to apply the connectivity mapping once, i.e., the 2nd parameter
                # direction has to become the 3rd
                n_apply_mapping = 1
            elif thickness_direction == 0:
                # We need to apply the connectivity mapping twice, i.e., the 1nd parameter
                # direction has to become the 3rd
                n_apply_mapping = 2

            # This permutes the parameter coordinate the following way:
            # [xi,eta,zeta]->[zeta,xi,eta]
            mapping = [2, 6, 7, 3, 1, 5, 4, 0]
            for _ in range(n_apply_mapping):
                element.nodes = [element.nodes[local_index] for local_index in mapping]


def get_visualization_third_parameter_direction_hex8(mesh):
    """Return a pyvista mesh with cell data for the third parameter direction
    for hex8 elements."""

    vtk_solid = mesh.get_vtk_representation()[1].grid
    pv_solid = _pv.UnstructuredGrid(vtk_solid)

    cell_thickness_direction = []
    for element in mesh.elements:
        if isinstance(element, _VolumeHEX8):
            _, jacobian_center = get_hex8_element_center_and_jacobian_mapping(element)
            cell_thickness_direction.append(jacobian_center[2])
        elif isinstance(element, _VolumeElement):
            cell_thickness_direction.append([0, 0, 0])

    if not len(cell_thickness_direction) == pv_solid.number_of_cells:
        raise ValueError(
            "Expected the same number of cells from the mesh and from the "
            f"pyvista object. Got {len(cell_thickness_direction)} form the "
            f"mesh and {pv_solid.number_of_cells} form pyvista"
        )

    pv_solid.cell_data["thickness_direction"] = cell_thickness_direction
    return pv_solid


def visualize_third_parameter_direction_hex8(mesh):
    """Visualize the third parameter direction for hex8 elements.

    This can be used to check the correct definition of the shell
    thickness for solid shell elements.
    """

    grid = get_visualization_third_parameter_direction_hex8(mesh)
    grid = grid.clean()
    cell_centers = grid.cell_centers()
    thickness_direction = cell_centers.glyph(
        orient="thickness_direction", scale="thickness_direction", factor=5
    )

    plotter = _pv.Plotter()
    plotter.renderer.add_axes()
    plotter.add_mesh(grid, color="white", show_edges=True, opacity=0.5)
    plotter.add_mesh(thickness_direction, color="red")
    plotter.show()



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Functions to create beam meshes along straight lines."""

import numpy as _np

from beamme.core.conf import bme as _bme
from beamme.core.rotation import Rotation as _Rotation
from beamme.mesh_creation_functions.beam_generic import (
    create_beam_mesh_generic as _create_beam_mesh_generic,
)


def create_beam_mesh_line(mesh, beam_class, material, start_point, end_point, **kwargs):
    """Generate a straight line of beam elements.

    Args
    ----
    mesh: Mesh
        Mesh that the line will be added to.
    beam_class: Beam
        Class of beam that will be used for this line.
    material: Material
        Material for this line.
    start_point, end_point: _np.array, list
        3D-coordinates for the start and end point of the line.

    **kwargs (for all of them look into create_beam_mesh_function)
    ----
    n_el: int
        Number of equally spaced beam elements along the line. Defaults to 1.
        Mutually exclusive with l_el.
    l_el: float
        Desired length of beam elements. Mutually exclusive with n_el.
        Be aware, that this length might not be achieved, if the elements are
        warped after they are created.
    start_node: Node, GeometrySet
        Node to use as the first node for this line. Use this if the line
        is connected to other lines (angles have to be the same, otherwise
        connections should be used). If a geometry set is given, it can
        contain one, and one node only.

    Return
    ----
    return_set: GeometryName
        Set with the 'start' and 'end' node of the line. Also a 'line' set
        with all nodes of the line.
    """

    # Get geometrical values for this line.
    start_point = _np.asarray(start_point)
    end_point = _np.asarray(end_point)
    direction = end_point - start_point
    line_length = _np.linalg.norm(direction)
    t1 = direction / line_length

    # Check if the z or y axis are larger projected onto the direction.
    # The tolerance is used here to ensure that round-off changes in the last digits of
    # the floating point values don't switch the case. This increases the robustness in
    # testing.
    if abs(_np.dot(t1, [0, 0, 1])) < abs(_np.dot(t1, [0, 1, 0])) - _bme.eps_quaternion:
        t2 = [0, 0, 1]
    else:
        t2 = [0, 1, 0]
    rotation = _Rotation.from_basis(t1, t2)

    def get_beam_geometry(parameter_a, parameter_b):
        """Return a function for the position along the beams axis."""

        def beam_function(xi):
            """Return a point on the beams axis for a given parameter
            coordinate xi."""
            point_a = start_point + parameter_a * direction
            point_b = start_point + parameter_b * direction
            pos = 0.5 * (1 - xi) * point_a + 0.5 * (1 + xi) * point_b
            arc_length = (
                0.5 * (1 - xi) * parameter_a + 0.5 * (1 + xi) * parameter_b
            ) * line_length
            return (pos, rotation, arc_length)

        return beam_function

    # Create the beam in the mesh
    return _create_beam_mesh_generic(
        mesh,
        beam_class=beam_class,
        material=material,
        function_generator=get_beam_geometry,
        interval=[0.0, 1.0],
        interval_length=line_length,
        **kwargs,
    )



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""Create a steel wire."""

import numpy as _np

from beamme.core.geometry_set import GeometryName as _GeometryName
from beamme.core.geometry_set import GeometrySet as _GeometrySet
from beamme.mesh_creation_functions.beam_line import (
    create_beam_mesh_line as _create_beam_mesh_line,
)
from beamme.utils.nodes import check_node_by_coordinate as _check_node_by_coordinate


def create_wire_fibers(
    mesh, beam_class, material, length, *, radius=None, layers=1, n_el=1
):
    """Create a steel wire consisting of multiple filaments. The wire will be
    oriented in x-direction.

    Args
    ----
    mesh: Mesh
        Mesh that the line will be added to.
    beam_class: Beam
        Class of beam that will be used for this line.
    material: Material
        Material for this line.
    length: float
        Length of the wire.
    radius: float
        If this parameter is given, not the beam cross section radius will be
        taken, but this one.
    layers: int
        Number of layers to be used for the wire.
    n_el: int
        Number of beam elements per filament.

    Return
    ----
    return_set: GeometryName
        Set with the 'start' and 'end' nodes of the wire. Also a 'all' set
        with all nodes of the wire.
    """

    if len(mesh.nodes) != 0:
        raise ValueError(
            "The create_wire_fibers function can only be used with an empty mesh."
        )

    if radius is None:
        wire_beam_radius = material.radius
    else:
        wire_beam_radius = radius

    def create_line(pos_yz):
        """Create a line starting at the yz-plane with the 2D coordinates
        pos_yz."""
        _create_beam_mesh_line(
            mesh,
            beam_class,
            material,
            [0.0, pos_yz[0], pos_yz[1]],
            [length, pos_yz[0], pos_yz[1]],
            n_el=n_el,
        )

    # Create the center filament.
    create_line([0.0, 0.0])

    # Create the filaments in the layers.
    for i_angle in range(6):
        angle = i_angle * _np.pi / 3.0
        direction_radial = _np.array([_np.cos(angle), _np.sin(angle)])
        angle = i_angle * _np.pi / 3.0 + 2.0 * _np.pi / 3.0
        direction_tangential = _np.array([_np.cos(angle), _np.sin(angle)])
        for i_layer in range(layers):
            for i_tangent in range(i_layer + 1):
                pos = (
                    2.0
                    * wire_beam_radius
                    * (
                        direction_radial * (i_layer + 1)
                        + direction_tangential * i_tangent
                    )
                )
                create_line(pos)

    # Create the sets to return.
    return_set = _GeometryName()
    start_nodes = mesh.get_nodes_by_function(_check_node_by_coordinate, 0, 0.0)
    end_nodes = mesh.get_nodes_by_function(_check_node_by_coordinate, 0, length)
    return_set["start"] = _GeometrySet(start_nodes)
    return_set["end"] = _GeometrySet(end_nodes)
    return_set["all"] = _GeometrySet(mesh.elements)
    return return_set



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This module implements a basic class to manage geometry in the input
file."""

from typing import KeysView as _KeysView
from typing import Sequence as _Sequence
from typing import Union as _Union
from typing import cast as _cast

import beamme.core.conf as _conf
from beamme.core.base_mesh_item import BaseMeshItem as _BaseMeshItem
from beamme.core.conf import bme as _bme
from beamme.core.container import ContainerBase as _ContainerBase
from beamme.core.element import Element as _Element
from beamme.core.element_beam import Beam as _Beam
from beamme.core.node import Node as _Node


class GeometrySetBase(_BaseMeshItem):
    """Base class for a geometry set."""

    def __init__(
        self, geometry_type: _conf.Geometry, name: str | None = None, **kwargs
    ):
        """Initialize the geometry set.

        Args:
            geometry_type: Type of geometry. Only geometry sets of a single specified geometry type are supported.
            name: Optional name to identify this geometry set.
        """
        super().__init__(**kwargs)

        self.geometry_type = geometry_type
        self.name = name

    def link_to_nodes(
        self, *, link_to_nodes: str = "explicitly_contained_nodes"
    ) -> None:
        """Set a link to this object in the all contained nodes of this
        geometry set.

        Args:
            link_to_nodes:
                "explicitly_contained_nodes":
                    A link will be set for all nodes that are explicitly part of the geometry set
                "all_nodes":
                    A link will be set for all nodes that are part of the geometry set, i.e., also
                    nodes connected to elements of an element set. This is mainly used for vtk
                    output so we can color the nodes which are part of element sets.
        """
        node_list: list[_Node] | _KeysView[_Node]
        if link_to_nodes == "explicitly_contained_nodes":
            node_list = self.get_node_dict().keys()
        elif link_to_nodes == "all_nodes":
            node_list = self.get_all_nodes()
        else:
            raise ValueError(f'Got unexpected value link nodes="{link_to_nodes}"')
        for node in node_list:
            node.node_sets_link.append(self)

    def check_replaced_nodes(self) -> None:
        """Check if nodes in this set have to be replaced.

        We need to do this for explicitly contained nodes in this set.
        """
        # Don't iterate directly over the keys as the dict changes during this iteration
        for node in list(self.get_node_dict().keys()):
            if node.master_node is not None:
                self.replace_node(node, node.get_master_node())

    def replace_node(self, old_node: _Node, new_node: _Node) -> None:
        """Replace an existing node in this geometry set with a new one.

        Args:
            old_node: Node to be replaced.
            new_node: Node that will be placed instead of old_node.
        """

        explicit_nodes_in_this_set = self.get_node_dict()
        explicit_nodes_in_this_set[new_node] = None
        del explicit_nodes_in_this_set[old_node]

    def get_node_dict(self) -> dict[_Node, None]:
        """Determine the explicitly added nodes for this set, i.e., nodes
        contained in elements are not returned.

        Returns:
            A dictionary containing the explicitly added nodes for this set.
        """
        raise NotImplementedError(
            'The "get_node_dict" method has to be overwritten in the derived class'
        )

    def get_points(self) -> list[_Node]:
        """Determine all points (represented by nodes) for this set.

        This function only works for point sets.

        Returns:
            A list containing the points (represented by nodes) associated with this set.
        """
        raise NotImplementedError(
            'The "get_points" method has to be overwritten in the derived class'
        )

    def get_all_nodes(self) -> list[_Node]:
        """Determine all nodes associated with this set.

        This includes nodes contained within the geometry added to this
        set, e.g., nodes connected to elements in element sets.

        Returns:
            A list containing all associated nodes.
        """
        raise NotImplementedError(
            'The "get_all_nodes" method has to be overwritten in the derived class'
        )

    def __add__(self, other):
        """Create a new geometry set with the combined geometries from this set
        and the other set.

        Args:
            other: Geometry set to be added to this one. This has to be of the same geometry type as this set.
        Returns:
            A combined geometry set.
        """
        combined_set = self.copy()
        combined_set.add(other)
        return combined_set


class GeometrySet(GeometrySetBase):
    """Geometry set which is defined by geometric entries."""

    def __init__(
        self,
        geometry: _Node | _Element | _Sequence[_Node | _Element] | "GeometrySet",
        **kwargs,
    ):
        """Initialize the geometry set.

        Args:
            geometry: Geometry entries to be contained in this set.
        """

        # This is ok, we check every single type in the add method
        if isinstance(geometry, list):
            geometry_type = self._get_geometry_type(geometry[0])
        else:
            geometry_type = self._get_geometry_type(geometry)

        super().__init__(geometry_type, **kwargs)

        self.geometry_objects: dict[_conf.Geometry, dict[_Node | _Element, None]] = {}
        for geo in _bme.geo:
            self.geometry_objects[geo] = {}
        self.add(geometry)

    @staticmethod
    def _get_geometry_type(
        item: _Node | _Element | _Sequence[_Node | _Element] | "GeometrySet",
    ) -> _conf.Geometry:
        """Get the geometry type of a given item.

        Returns:
            Geometry type of the geometry set.
        """

        if isinstance(item, _Node):
            return _bme.geo.point
        elif isinstance(item, _Beam):
            return _bme.geo.line
        elif isinstance(item, GeometrySet):
            return item.geometry_type
        raise TypeError(f"Got unexpected type {type(item)}")

    def add(
        self, item: _Node | _Element | _Sequence[_Node | _Element] | "GeometrySet"
    ) -> None:
        """Add geometry item(s) to this object."""

        if isinstance(item, list):
            for sub_item in item:
                self.add(sub_item)
        elif isinstance(item, GeometrySet):
            if item.geometry_type is self.geometry_type:
                for geometry in item.geometry_objects[self.geometry_type]:
                    self.add(geometry)
            else:
                raise TypeError(
                    "You tried to add a {item.geometry_type} set to a {self.geometry_type} set. "
                    "This is not possible"
                )
        elif self._get_geometry_type(item) is self.geometry_type:
            self.geometry_objects[self.geometry_type][_cast(_Node | _Element, item)] = (
                None
            )
        else:
            raise TypeError(f"Got unexpected geometry type {type(item)}")

    def get_node_dict(self) -> dict[_Node, None]:
        """Determine the explicitly added nodes for this set, i.e., nodes
        contained in elements for element sets are not returned.

        Thus, for non-point sets an empty dict is returned.

        Returns:
            A dictionary containing the explicitly added nodes for this set.
        """
        if self.geometry_type is _bme.geo.point:
            return _cast(dict[_Node, None], self.geometry_objects[_bme.geo.point])
        else:
            return {}

    def get_points(self) -> list[_Node]:
        """Determine all points (represented by nodes) for this set.

        This function only works for point sets.

        Returns:
            A list containing the points (represented by nodes) associated with this set.
        """
        if self.geometry_type is _bme.geo.point:
            return list(self.get_node_dict().keys())
        else:
            raise TypeError(
                "The function get_points can only be called for point sets."
                f" The present type is {self.geometry_type}"
            )

    def get_all_nodes(self) -> list[_Node]:
        """Determine all nodes associated with this set.

        This includes nodes contained within the geometry added to this
        set, e.g., nodes connected to elements in element sets.

        Returns:
            A list containing all associated nodes.
        """

        if self.geometry_type is _bme.geo.point:
            return list(
                _cast(_KeysView[_Node], self.geometry_objects[_bme.geo.point].keys())
            )
        elif self.geometry_type is _bme.geo.line:
            nodes = []
            for element in _cast(
                _KeysView[_Element], self.geometry_objects[_bme.geo.line].keys()
            ):
                nodes.extend(element.nodes)
            # Remove duplicates while preserving order
            return list(dict.fromkeys(nodes))
        else:
            raise TypeError(
                "Currently GeometrySet is only implemented for points and lines"
            )

    def get_geometry_objects(self) -> _Sequence[_Node | _Element]:
        """Get a list of the objects with the specified geometry type.

        Returns:
            A list with the contained geometry.
        """
        return list(self.geometry_objects[self.geometry_type].keys())

    def copy(self) -> "GeometrySet":
        """Create a shallow copy of this object, the reference to the nodes
        will be the same, but the containers storing them will be copied.

        Returns:
            A shallow copy of the geometry set.
        """
        return GeometrySet(list(self.geometry_objects[self.geometry_type].keys()))


class GeometrySetNodes(GeometrySetBase):
    """Geometry set which is defined by nodes and not explicit geometry."""

    def __init__(
        self,
        geometry_type: _conf.Geometry,
        nodes: _Union[_Node, list[_Node], "GeometrySetNodes", None] = None,
        **kwargs,
    ):
        """Initialize the geometry set.

        Args:
            geometry_type: Type of geometry. This is  necessary, as the boundary conditions
                and input file depend on that type.
            nodes: Node(s) or list of nodes to be added to this geometry set.
        """

        if geometry_type not in _bme.geo:
            raise TypeError(f"Expected geometry enum, got {geometry_type}")

        super().__init__(geometry_type, **kwargs)
        self.nodes: dict[_Node, None] = {}
        if nodes is not None:
            self.add(nodes)

    def add(self, value: _Union[_Node, list[_Node], "GeometrySetNodes"]) -> None:
        """Add nodes to this object.

        Args:
            nodes: Node(s) or list of nodes to be added to this geometry set.
        """

        if isinstance(value, list):
            # Loop over items and check if they are either Nodes or integers.
            # This improves the performance considerably when large list of
            # Nodes are added.
            for item in value:
                self.add(item)
        elif isinstance(value, (int, _Node)):
            self.nodes[value] = None
        elif isinstance(value, GeometrySetNodes):
            # Add all nodes from this geometry set.
            if self.geometry_type == value.geometry_type:
                for node in value.nodes:
                    self.add(node)
            else:
                raise TypeError(
                    f"You tried to add a {value.geometry_type} set to a {self.geometry_type} set. "
                    "This is not possible"
                )
        else:
            raise TypeError(f"Expected Node or list, but got {type(value)}")

    def get_node_dict(self) -> dict[_Node, None]:
        """Determine the explicitly added nodes for this set.

        Thus, we can simply return all points here.

        Returns:
            A dictionary containing the explicitly added nodes for this set.
        """
        return self.nodes

    def get_points(self) -> list[_Node]:
        """Determine all points (represented by nodes) for this set.

        This function only works for point sets.

        Returns:
            A list containing the points (represented by nodes) associated with this set.
        """
        if self.geometry_type is _bme.geo.point:
            return list(self.get_node_dict().keys())
        else:
            raise TypeError(
                "The function get_points can only be called for point sets."
                f" The present type is {self.geometry_type}"
            )

    def get_all_nodes(self) -> list[_Node]:
        """Determine all nodes associated with this set.

        This includes nodes contained within the geometry added to this
        set, e.g., nodes connected to elements in element sets.

        Returns:
            A list containing all associated nodes.
        """
        return list(self.get_node_dict().keys())

    def copy(self) -> "GeometrySetNodes":
        """Create a shallow copy of this object, the reference to the nodes
        will be the same, but the containers storing them will be copied.

        Returns:
            A shallow copy of the geometry set.
        """
        return GeometrySetNodes(
            geometry_type=self.geometry_type,
            nodes=list(self.nodes.keys()),
        )


class GeometryName(dict):
    """Group node geometry sets together.

    This is mainly used for export from mesh functions. The sets can be
    accessed by a unique name. There is no distinction between different
    types of geometry, every name can only be used once -> use
    meaningful names.
    """

    def __setitem__(self, key, value):
        """Set a geometry set in this container."""

        if not isinstance(key, str):
            raise TypeError(f"Expected string, got {type(key)}!")
        if isinstance(value, GeometrySetBase):
            super().__setitem__(key, value)
        else:
            raise NotImplementedError("GeometryName can only store GeometrySets")


class GeometrySetContainer(_ContainerBase):
    """A class to group geometry sets together with the key being the geometry
    type."""

    def __init__(self, *args, **kwargs):
        """Initialize the container and create the default keys in the map."""
        super().__init__(*args, **kwargs)

        self.item_types = [GeometrySetBase]

        for geometry_key in _bme.geo:
            self[geometry_key] = []

    def copy(self):
        """When creating a copy of this object, all lists in this object will
        be copied also."""

        # Create a new geometry set container.
        copy = GeometrySetContainer()

        # Add a copy of every list from this container to the new one.
        for geometry_key in _bme.geo:
            copy[geometry_key] = self[geometry_key].copy()

        return copy



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# general configuration
project = "BeamMe"
copyright = "2026, BeamMe Authors"
author = "BeamMe Authors"

# html theme configuration
html_theme = "pydata_sphinx_theme"
html_title = "BeamMe"
html_theme_options = {
    "github_url": "https://github.com/beamme-py/beamme",
}
html_show_sourcelink = False  # Hide "View Source" link on the right side of the page

# extensions
extensions = [
    "myst_parser",  # to enable Markdown support
    "sphinxcontrib.jquery",  # to enable custom JavaScript (open links in new empty tab)
]

# markdown configuration
myst_enable_extensions = [
    "colon_fence",  # For ::: fenced code blocks
    "linkify",  # Auto-detects URLs and makes them hyperlinks
]
myst_heading_anchors = 3  # automatic heading anchors for Markdown files

# JavaScript configuration (to open links in new tabs)
html_js_files = ["js/custom.js"]
html_static_path = ["static"]



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file defines the interface to the ArborX geometric search
functionality."""

from beamme.geometric_search.utils import arborx_is_available as _arborx_is_available

if _arborx_is_available():
    from beamme.geometric_search.arborx_lib import (
        find_close_points as _find_close_points,
    )
    from beamme.geometric_search.arborx_lib import (
        kokkos_finalize as _kokkos_finalize,
    )
    from beamme.geometric_search.arborx_lib import (
        kokkos_initialize as _kokkos_initialize,
    )


class KokkosScopeGuardWrapper:
    """Wrap the initialize and finalize calls to Kokkos."""

    def __init__(self):
        """Call initialize when this object is created."""
        _kokkos_initialize()

    def __del__(self):
        """Finalize Kokkos after this object goes out of scope, i.e., at the
        end of this modules lifetime."""
        _kokkos_finalize()


if _arborx_is_available():
    # Create the scope guard
    kokkos_scope_guard_wrapper = KokkosScopeGuardWrapper()


def find_close_points_arborx(point_coordinates, tol):
    """Call the ArborX implementation of find close_points."""
    if _arborx_is_available():
        return _find_close_points(point_coordinates, tol)
    else:
        raise ModuleNotFoundError("ArborX functionality is not available")



# The MIT License (MIT)
#
# Copyright (c) 2018-2026 BeamMe Authors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""This file defines the interface to the Cython geometric search
functionality."""

import warnings as _warnings

from beamme.geometric_search.utils import cython_is_available as _cython_is_available

if _cython_is_available():
    from beamme.geometric_search.cython_lib import (
        find_close_points as _find_close_points,
    )


def find_close_points_brute_force_cython(
    point_coordinates, tol, *, n_points_performance_warning=5000
):
    """Call the Cython brute force implementation of find close_points."""
    if _cython_is_available():
        n_points = len(point_coordinates)
        if n_points > n_points_performance_warning:
            _warnings.warn(
                "The function find_close_points is called with the brute force algorithm "
                + f"with {n_points} points, for performance reasons other algorithms should be used!"
            )
        return _find_close_points(point_coordinates, tol)
    else:
        raise ModuleNotFoundError(
            "Cython geometric search functionality is not available"
        )
